#include "tiles_generic.h"
#include "z80.h"
#include "namco_snd.h"

static unsigned char DrvInputPort0[8] = {0, 0, 0, 0, 0, 0, 0, 0};
static unsigned char DrvInputPort1[8] = {0, 0, 0, 0, 0, 0, 0, 0};
static unsigned char DrvInputPort2[8] = {0, 0, 0, 0, 0, 0, 0, 0};
static unsigned char DrvDip[3]        = {0, 0, 0};
static unsigned char DrvInput[3]      = {0x00, 0x00, 0x00};
static unsigned char DrvReset         = 0;

static unsigned char *Mem                 = NULL;
static unsigned char *MemEnd              = NULL;
static unsigned char *RamStart            = NULL;
static unsigned char *RamEnd              = NULL;
static unsigned char *DrvZ80Rom1          = NULL;
static unsigned char *DrvZ80Rom2          = NULL;
static unsigned char *DrvZ80Rom3          = NULL;
static unsigned char *DrvVideoRam         = NULL;
static unsigned char *DrvSharedRam1       = NULL;
static unsigned char *DrvSharedRam2       = NULL;
static unsigned char *DrvSharedRam3       = NULL;
static unsigned char *DrvPromPalette      = NULL;
static unsigned char *DrvPromCharLookup   = NULL;
static unsigned char *DrvPromSpriteLookup = NULL;
static unsigned char *DrvChars            = NULL;
static unsigned char *DrvSprites          = NULL;
static unsigned char *DrvTempRom          = NULL;
static unsigned int  *DrvPalette          = NULL;

static unsigned char DrvCPU1FireIRQ;
static unsigned char DrvCPU2FireIRQ;
static unsigned char DrvCPU3FireIRQ;
static unsigned char DrvCPU2Halt;
static unsigned char DrvCPU3Halt;
static unsigned char DrvFlipScreen;
static unsigned char DrvStarControl[6];
static unsigned int  DrvStarScrollX;
static unsigned int  DrvStarScrollY;

static unsigned char IOChipCustomCommand;
static unsigned char IOChipCPU1FireIRQ;
static unsigned char IOChipMode;
static unsigned char IOChipCredits;
static unsigned char IOChipCoinPerCredit;
static unsigned char IOChipCreditPerCoin;
static unsigned char IOChipCustom[16];
static unsigned char PrevInValue;

static int nCyclesDone[3], nCyclesTotal[3];
static int nCyclesSegment;

static Z80_Regs Z80_0;
static Z80_Regs Z80_1;
static Z80_Regs Z80_2;

struct CPU_Config {
	Z80ReadIoHandler Z80In;
	Z80WriteIoHandler Z80Out;
	Z80ReadProgHandler Z80Read;
	Z80WriteProgHandler Z80Write;
	Z80ReadOpHandler Z80ReadOp;
	Z80ReadOpArgHandler Z80ReadOpArg;
};

static int nActiveCPU;
static CPU_Config Z80_0_Config;
static CPU_Config Z80_1_Config;
static CPU_Config Z80_2_Config;

static struct BurnInputInfo DrvInputList[] =
{
	{"Coin 1"            , BIT_DIGITAL  , DrvInputPort0 + 4, "p1 coin"   },
	{"Start 1"           , BIT_DIGITAL  , DrvInputPort0 + 2, "p1 start"  },
	{"Coin 2"            , BIT_DIGITAL  , DrvInputPort0 + 5, "p2 coin"   },
	{"Start 2"           , BIT_DIGITAL  , DrvInputPort0 + 3, "p2 start"  },

	{"Left"              , BIT_DIGITAL  , DrvInputPort1 + 3, "p1 left"   },
	{"Right"             , BIT_DIGITAL  , DrvInputPort1 + 1, "p1 right"  },
	{"Fire 1"            , BIT_DIGITAL  , DrvInputPort1 + 4, "p1 fire 1" },
	
	{"Left (Cocktail)"   , BIT_DIGITAL  , DrvInputPort2 + 3, "p2 left"   },
	{"Right (Cocktail)"  , BIT_DIGITAL  , DrvInputPort2 + 1, "p2 right"  },
	{"Fire 1 (Cocktail)" , BIT_DIGITAL  , DrvInputPort2 + 4, "p2 fire 1" },

	{"Reset"             , BIT_DIGITAL  , &DrvReset        , "reset"     },
	{"Service"           , BIT_DIGITAL  , DrvInputPort0 + 6, "service"   },
	{"Dip 1"             , BIT_DIPSWITCH, DrvDip + 0       , "dip"       },
	{"Dip 2"             , BIT_DIPSWITCH, DrvDip + 1       , "dip"       },
	{"Dip 3"             , BIT_DIPSWITCH, DrvDip + 2       , "dip"       },
};

STDINPUTINFO(Drv);

static inline void DrvMakeInputs()
{
	// Reset Inputs
	DrvInput[0] = 0x7f;
	DrvInput[1] = 0xff;
	DrvInput[2] = 0xff;

	// Compile Digital Inputs
	for (int i = 0; i < 8; i++) {
		DrvInput[0] -= (DrvInputPort0[i] & 1) << i;
		DrvInput[1] -= (DrvInputPort1[i] & 1) << i;
		DrvInput[2] -= (DrvInputPort2[i] & 1) << i;
	}
}

static struct BurnDIPInfo DrvDIPList[]=
{
	// Default Values
	{0x0c, 0xff, 0xff, 0x80, NULL                     },
	{0x0d, 0xff, 0xff, 0xf7, NULL                     },
	{0x0e, 0xff, 0xff, 0x97, NULL                     },
	
	// Dip 1
	{0   , 0xfe, 0   , 2   , "Service Mode"           },
	{0x0c, 0x01, 0x80, 0x80, "Off"                    },
	{0x0c, 0x01, 0x80, 0x00, "On"                     },

	// Dip 2
	{0   , 0xfe, 0   , 4   , "Difficulty"             },
	{0x0d, 0x01, 0x03, 0x03, "Easy"                   },
	{0x0d, 0x01, 0x03, 0x00, "Medium"                 },
	{0x0d, 0x01, 0x03, 0x01, "Hard"                   },
	{0x0d, 0x01, 0x03, 0x02, "Hardest"                },
	
	{0   , 0xfe, 0   , 2   , "Demo Sounds"            },
	{0x0d, 0x01, 0x08, 0x08, "Off"                    },
	{0x0d, 0x01, 0x08, 0x00, "On"                     },
	
	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x0d, 0x01, 0x10, 0x10, "Off"                    },
	{0x0d, 0x01, 0x10, 0x00, "On"                     },
	
	{0   , 0xfe, 0   , 2   , "Rack Test"              },
	{0x0d, 0x01, 0x20, 0x20, "Off"                    },
	{0x0d, 0x01, 0x20, 0x00, "On"                     },
	
	{0   , 0xfe, 0   , 2   , "Cabinet"                },
	{0x0d, 0x01, 0x80, 0x80, "Upright"                },
	{0x0d, 0x01, 0x80, 0x00, "Cocktail"               },
	
	// Dip 3	
	{0   , 0xfe, 0   , 8   , "Coinage"                },
	{0x0e, 0x01, 0x07, 0x04, "4 Coins 1 Play"         },
	{0x0e, 0x01, 0x07, 0x02, "3 Coins 1 Play"         },
	{0x0e, 0x01, 0x07, 0x06, "2 Coins 1 Play"         },
	{0x0e, 0x01, 0x07, 0x07, "1 Coin  1 Play"         },
	{0x0e, 0x01, 0x07, 0x01, "2 Coins 3 Plays"        },
	{0x0e, 0x01, 0x07, 0x03, "1 Coin  2 Plays"        },
	{0x0e, 0x01, 0x07, 0x05, "1 Coin  3 Plays"        },
	{0x0e, 0x01, 0x07, 0x00, "Freeplay"               },	
	
	{0   , 0xfe, 0   , 8   , "Bonus Life"             },
	{0x0e, 0x01, 0x38, 0x20, "20k  60k  60k"          },
	{0x0e, 0x01, 0x38, 0x18, "20k  60k"               },
	{0x0e, 0x01, 0x38, 0x10, "20k  70k  70k"          },
	{0x0e, 0x01, 0x38, 0x30, "20k  80k  80k"          },
	{0x0e, 0x01, 0x38, 0x38, "30k  80k"               },
	{0x0e, 0x01, 0x38, 0x08, "30k 100k 100k"          },
	{0x0e, 0x01, 0x38, 0x28, "30k 120k 120k"          },
	{0x0e, 0x01, 0x38, 0x00, "None"                   },	
	
	{0   , 0xfe, 0   , 4   , "Lives"                  },
	{0x0e, 0x01, 0xc0, 0x00, "2"                      },
	{0x0e, 0x01, 0xc0, 0x80, "3"                      },
	{0x0e, 0x01, 0xc0, 0x40, "4"                      },
	{0x0e, 0x01, 0xc0, 0xc0, "5"                      },
};

STDDIPINFO(Drv);

static struct BurnDIPInfo GalagamwDIPList[]=
{
	// Default Values
	{0x0c, 0xff, 0xff, 0x80, NULL                     },
	{0x0d, 0xff, 0xff, 0xf7, NULL                     },
	{0x0e, 0xff, 0xff, 0x97, NULL                     },
	
	// Dip 1
	{0   , 0xfe, 0   , 2   , "Service Mode"           },
	{0x0c, 0x01, 0x80, 0x80, "Off"                    },
	{0x0c, 0x01, 0x80, 0x00, "On"                     },

	// Dip 2
	{0   , 0xfe, 0   , 2   , "2 Credits Game"         },
	{0x0d, 0x01, 0x01, 0x00, "1 Player"               },
	{0x0d, 0x01, 0x01, 0x01, "2 Players"              },
	
	{0   , 0xfe, 0   , 4   , "Difficulty"             },
	{0x0d, 0x01, 0x06, 0x06, "Easy"                   },
	{0x0d, 0x01, 0x06, 0x00, "Medium"                 },
	{0x0d, 0x01, 0x06, 0x02, "Hard"                   },
	{0x0d, 0x01, 0x06, 0x04, "Hardest"                },
	
	{0   , 0xfe, 0   , 2   , "Demo Sounds"            },
	{0x0d, 0x01, 0x08, 0x08, "Off"                    },
	{0x0d, 0x01, 0x08, 0x00, "On"                     },
	
	{0   , 0xfe, 0   , 2   , "Freeze"                 },
	{0x0d, 0x01, 0x10, 0x10, "Off"                    },
	{0x0d, 0x01, 0x10, 0x00, "On"                     },
	
	{0   , 0xfe, 0   , 2   , "Rack Test"              },
	{0x0d, 0x01, 0x20, 0x20, "Off"                    },
	{0x0d, 0x01, 0x20, 0x00, "On"                     },
	
	{0   , 0xfe, 0   , 2   , "Cabinet"                },
	{0x0d, 0x01, 0x80, 0x80, "Upright"                },
	{0x0d, 0x01, 0x80, 0x00, "Cocktail"               },
	
	// Dip 3	
	{0   , 0xfe, 0   , 8   , "Coinage"                },
	{0x0e, 0x01, 0x07, 0x04, "4 Coins 1 Play"         },
	{0x0e, 0x01, 0x07, 0x02, "3 Coins 1 Play"         },
	{0x0e, 0x01, 0x07, 0x06, "2 Coins 1 Play"         },
	{0x0e, 0x01, 0x07, 0x07, "1 Coin  1 Play"         },
	{0x0e, 0x01, 0x07, 0x01, "2 Coins 3 Plays"        },
	{0x0e, 0x01, 0x07, 0x03, "1 Coin  2 Plays"        },
	{0x0e, 0x01, 0x07, 0x05, "1 Coin  3 Plays"        },
	{0x0e, 0x01, 0x07, 0x00, "Freeplay"               },	
	
	{0   , 0xfe, 0   , 8   , "Bonus Life"             },
	{0x0e, 0x01, 0x38, 0x20, "20k  60k  60k"          },
	{0x0e, 0x01, 0x38, 0x18, "20k  60k"               },
	{0x0e, 0x01, 0x38, 0x10, "20k  70k  70k"          },
	{0x0e, 0x01, 0x38, 0x30, "20k  80k  80k"          },
	{0x0e, 0x01, 0x38, 0x38, "30k  80k"               },
	{0x0e, 0x01, 0x38, 0x08, "30k 100k 100k"          },
	{0x0e, 0x01, 0x38, 0x28, "30k 120k 120k"          },
	{0x0e, 0x01, 0x38, 0x00, "None"                   },	
	
	{0   , 0xfe, 0   , 4   , "Lives"                  },
	{0x0e, 0x01, 0xc0, 0x00, "2"                      },
	{0x0e, 0x01, 0xc0, 0x80, "3"                      },
	{0x0e, 0x01, 0xc0, 0x40, "4"                      },
	{0x0e, 0x01, 0xc0, 0xc0, "5"                      },
};

STDDIPINFO(Galagamw);

static struct BurnRomInfo DrvRomDesc[] = {
	{ "gg1-1b.3p",     0x01000, 0xab036c9f, BRF_ESS | BRF_PRG }, //  0	Z80 #1 Program Code
	{ "gg1-2b.3m",     0x01000, 0xd9232240, BRF_ESS | BRF_PRG }, //	 1
	{ "gg1-3.2m",      0x01000, 0x753ce503, BRF_ESS | BRF_PRG }, //	 2
	{ "gg1-4b.2l",     0x01000, 0x499fcc76, BRF_ESS | BRF_PRG }, //	 3
	
	{ "gg1-5b.3f",     0x01000, 0xbb5caae3, BRF_ESS | BRF_PRG }, //  4	Z80 #2 Program Code
	
	{ "gg1-7b.2c",     0x01000, 0xd016686b, BRF_ESS | BRF_PRG }, //  5	Z80 #3 Program Code
	
	{ "54xx.bin",      0x00400, 0xee7357e0, BRF_ESS | BRF_PRG }, //  6	54XX Program Code
	
	{ "51xx.bin",      0x00400, 0xc2f57ef8, BRF_ESS | BRF_PRG }, //  7	51XX Program Code
	
	{ "gg1-9.4l",      0x01000, 0x58b2f47c, BRF_GRA },	     //  8	Characters
	
	{ "gg1-11.4d",     0x01000, 0xad447c80, BRF_GRA },	     //  9	Sprites
	{ "gg1-10.4f",     0x01000, 0xdd6f1afc, BRF_GRA },	     //  10
	
	{ "prom-5.5n",     0x00020, 0x54603c6b, BRF_GRA },	     //  11	PROMs
	{ "prom-4.2n",     0x00100, 0x59b6edab, BRF_GRA },	     //  12
	{ "prom-3.1c",     0x00100, 0x4a04bb6b, BRF_GRA },	     //  13
	{ "prom-1.1d",     0x00100, 0x7a2815b4, BRF_GRA },	     //  14
	{ "prom-2.5c",     0x00100, 0x77245b66, BRF_GRA },	     //  15
};

STD_ROM_PICK(Drv);
STD_ROM_FN(Drv);

static struct BurnRomInfo GalagaoRomDesc[] = {
	{ "gg1-1",         0x01000, 0xa3a0f743, BRF_ESS | BRF_PRG }, //  0	Z80 #1 Program Code
	{ "gg1-2",         0x01000, 0x43bb0d5c, BRF_ESS | BRF_PRG }, //	 1
	{ "gg1-3.2m",      0x01000, 0x753ce503, BRF_ESS | BRF_PRG }, //	 2
	{ "gg1-4",         0x01000, 0x83874442, BRF_ESS | BRF_PRG }, //	 3
	
	{ "gg1-5",         0x01000, 0x3102fccd, BRF_ESS | BRF_PRG }, //  4	Z80 #2 Program Code
	
	{ "gg1-7",         0x01000, 0x8995088d, BRF_ESS | BRF_PRG }, //  5	Z80 #3 Program Code
	
	{ "54xx.bin",      0x00400, 0xee7357e0, BRF_ESS | BRF_PRG }, //  6	54XX Program Code
	
	{ "51xx.bin",      0x00400, 0xc2f57ef8, BRF_ESS | BRF_PRG }, //  7	51XX Program Code
	
	{ "gg1-9.4l",      0x01000, 0x58b2f47c, BRF_GRA },	     //  8	Characters
	
	{ "gg1-11.4d",     0x01000, 0xad447c80, BRF_GRA },	     //  9	Sprites
	{ "gg1-10.4f",     0x01000, 0xdd6f1afc, BRF_GRA },	     //  10
	
	{ "prom-5.5n",     0x00020, 0x54603c6b, BRF_GRA },	     //  11	PROMs
	{ "prom-4.2n",     0x00100, 0x59b6edab, BRF_GRA },	     //  12
	{ "prom-3.1c",     0x00100, 0x4a04bb6b, BRF_GRA },	     //  13
	{ "prom-1.1d",     0x00100, 0x7a2815b4, BRF_GRA },	     //  14
	{ "prom-2.5c",     0x00100, 0x77245b66, BRF_GRA },	     //  15
};

STD_ROM_PICK(Galagao);
STD_ROM_FN(Galagao);

static struct BurnRomInfo GalagamwRomDesc[] = {
	{ "3200a.bin",     0x01000, 0x3ef0b053, BRF_ESS | BRF_PRG }, //  0	Z80 #1 Program Code
	{ "3300b.bin",     0x01000, 0x1b280831, BRF_ESS | BRF_PRG }, //	 1
	{ "3400c.bin",     0x01000, 0x16233d33, BRF_ESS | BRF_PRG }, //	 2
	{ "3500d.bin",     0x01000, 0x0aaf5c23, BRF_ESS | BRF_PRG }, //	 3
	
	{ "3600e.bin",     0x01000, 0xbc556e76, BRF_ESS | BRF_PRG }, //  4	Z80 #2 Program Code
	
	{ "3700g.bin",     0x01000, 0xb07f0aa4, BRF_ESS | BRF_PRG }, //  5	Z80 #3 Program Code
	
	{ "54xx.bin",      0x00400, 0xee7357e0, BRF_ESS | BRF_PRG }, //  6	54XX Program Code
	
	{ "51xx.bin",      0x00400, 0xc2f57ef8, BRF_ESS | BRF_PRG }, //  7	51XX Program Code
	
	{ "gg1-9.4l",      0x01000, 0x58b2f47c, BRF_GRA },	     //  8	Characters
	
	{ "gg1-11.4d",     0x01000, 0xad447c80, BRF_GRA },	     //  9	Sprites
	{ "gg1-10.4f",     0x01000, 0xdd6f1afc, BRF_GRA },	     //  10
	
	{ "prom-5.5n",     0x00020, 0x54603c6b, BRF_GRA },	     //  11	PROMs
	{ "prom-4.2n",     0x00100, 0x59b6edab, BRF_GRA },	     //  12
	{ "prom-3.1c",     0x00100, 0x4a04bb6b, BRF_GRA },	     //  13
	{ "prom-1.1d",     0x00100, 0x7a2815b4, BRF_GRA },	     //  14
	{ "prom-2.5c",     0x00100, 0x77245b66, BRF_GRA },	     //  15
};

STD_ROM_PICK(Galagamw);
STD_ROM_FN(Galagamw);

static struct BurnRomInfo GalagamfRomDesc[] = {
	{ "3200a.bin",     0x01000, 0x3ef0b053, BRF_ESS | BRF_PRG }, //  0	Z80 #1 Program Code
	{ "3300b.bin",     0x01000, 0x1b280831, BRF_ESS | BRF_PRG }, //	 1
	{ "3400c.bin",     0x01000, 0x16233d33, BRF_ESS | BRF_PRG }, //	 2
	{ "3500d.bin",     0x01000, 0x0aaf5c23, BRF_ESS | BRF_PRG }, //	 3
	
	{ "3600fast.bin",  0x01000, 0x23d586e5, BRF_ESS | BRF_PRG }, //  4	Z80 #2 Program Code
	
	{ "3700g.bin",     0x01000, 0xb07f0aa4, BRF_ESS | BRF_PRG }, //  5	Z80 #3 Program Code
	
	{ "54xx.bin",      0x00400, 0xee7357e0, BRF_ESS | BRF_PRG }, //  6	54XX Program Code
	
	{ "51xx.bin",      0x00400, 0xc2f57ef8, BRF_ESS | BRF_PRG }, //  7	51XX Program Code
	
	{ "gg1-9.4l",      0x01000, 0x58b2f47c, BRF_GRA },	     //  8	Characters
	
	{ "gg1-11.4d",     0x01000, 0xad447c80, BRF_GRA },	     //  9	Sprites
	{ "gg1-10.4f",     0x01000, 0xdd6f1afc, BRF_GRA },	     //  10
	
	{ "prom-5.5n",     0x00020, 0x54603c6b, BRF_GRA },	     //  11	PROMs
	{ "prom-4.2n",     0x00100, 0x59b6edab, BRF_GRA },	     //  12
	{ "prom-3.1c",     0x00100, 0x4a04bb6b, BRF_GRA },	     //  13
	{ "prom-1.1d",     0x00100, 0x7a2815b4, BRF_GRA },	     //  14
	{ "prom-2.5c",     0x00100, 0x77245b66, BRF_GRA },	     //  15
};

STD_ROM_PICK(Galagamf);
STD_ROM_FN(Galagamf);

static struct BurnRomInfo GalagamkRomDesc[] = {
	{ "mk2-1",         0x01000, 0x23cea1e2, BRF_ESS | BRF_PRG }, //  0	Z80 #1 Program Code
	{ "mk2-2",         0x01000, 0x89695b1a, BRF_ESS | BRF_PRG }, //	 1
	{ "3400c.bin",     0x01000, 0x16233d33, BRF_ESS | BRF_PRG }, //	 2
	{ "mk2-4",         0x01000, 0x24b767f5, BRF_ESS | BRF_PRG }, //	 3
	
	{ "gg1-5",         0x01000, 0x3102fccd, BRF_ESS | BRF_PRG }, //  4	Z80 #2 Program Code
	
	{ "gg1-7b.2c",     0x01000, 0xd016686b, BRF_ESS | BRF_PRG }, //  5	Z80 #3 Program Code
	
	{ "54xx.bin",      0x00400, 0xee7357e0, BRF_ESS | BRF_PRG }, //  6	54XX Program Code
	
	{ "51xx.bin",      0x00400, 0xc2f57ef8, BRF_ESS | BRF_PRG }, //  7	51XX Program Code
	
	{ "gg1-9.4l",      0x01000, 0x58b2f47c, BRF_GRA },	     //  8	Characters
	
	{ "gg1-11.4d",     0x01000, 0xad447c80, BRF_GRA },	     //  9	Sprites
	{ "gg1-10.4f",     0x01000, 0xdd6f1afc, BRF_GRA },	     //  10
	
	{ "prom-5.5n",     0x00020, 0x54603c6b, BRF_GRA },	     //  11	PROMs
	{ "prom-4.2n",     0x00100, 0x59b6edab, BRF_GRA },	     //  12
	{ "prom-3.1c",     0x00100, 0x4a04bb6b, BRF_GRA },	     //  13
	{ "prom-1.1d",     0x00100, 0x7a2815b4, BRF_GRA },	     //  14
	{ "prom-2.5c",     0x00100, 0x77245b66, BRF_GRA },	     //  15
};

STD_ROM_PICK(Galagamk);
STD_ROM_FN(Galagamk);

static struct BurnRomInfo GallagRomDesc[] = {
	{ "gg1-1",         0x01000, 0xa3a0f743, BRF_ESS | BRF_PRG }, //  0	Z80 #1 Program Code
	{ "gallag.2",      0x01000, 0x5eda60a7, BRF_ESS | BRF_PRG }, //	 1
	{ "gg1-3.2m",      0x01000, 0x753ce503, BRF_ESS | BRF_PRG }, //	 2
	{ "gg1-4",         0x01000, 0x83874442, BRF_ESS | BRF_PRG }, //	 3
	
	{ "gg1-5",         0x01000, 0x3102fccd, BRF_ESS | BRF_PRG }, //  4	Z80 #2 Program Code
	
	{ "gg1-7",         0x01000, 0x8995088d, BRF_ESS | BRF_PRG }, //  5	Z80 #3 Program Code
	
	{ "gallag.6",      0x01000, 0x001b70bc, BRF_ESS | BRF_PRG }, //  6	Z80 #4 Program Code
	
	{ "gallag.8",      0x01000, 0x169a98a4, BRF_GRA },	     //  7	Characters
	
	{ "gg1-11.4d",     0x01000, 0xad447c80, BRF_GRA },	     //  8	Sprites
	{ "gg1-10.4f",     0x01000, 0xdd6f1afc, BRF_GRA },	     //  9
	
	{ "prom-5.5n",     0x00020, 0x54603c6b, BRF_GRA },	     //  10	PROMs
	{ "prom-4.2n",     0x00100, 0x59b6edab, BRF_GRA },	     //  11
	{ "prom-3.1c",     0x00100, 0x4a04bb6b, BRF_GRA },	     //  12
	{ "prom-1.1d",     0x00100, 0x7a2815b4, BRF_GRA },	     //  13
	{ "prom-2.5c",     0x00100, 0x77245b66, BRF_GRA },	     //  14
};

STD_ROM_PICK(Gallag);
STD_ROM_FN(Gallag);

static int MemIndex()
{
	unsigned char *Next; Next = Mem;

	DrvZ80Rom1             = Next; Next += 0x04000;
	DrvZ80Rom2             = Next; Next += 0x04000;
	DrvZ80Rom3             = Next; Next += 0x04000;
	DrvPromPalette         = Next; Next += 0x00020;
	DrvPromCharLookup      = Next; Next += 0x00100;
	DrvPromSpriteLookup    = Next; Next += 0x00100;
	NamcoSoundProm         = Next; Next += 0x00100;
	
	RamStart               = Next;

	DrvVideoRam            = Next; Next += 0x00800;
	DrvSharedRam1          = Next; Next += 0x00400;
	DrvSharedRam1          = Next; Next += 0x04000;
	DrvSharedRam2          = Next; Next += 0x00400;
	DrvSharedRam3          = Next; Next += 0x00400;

	RamEnd                 = Next;

	DrvChars               = Next; Next += 0x100 * 8 * 8;
	DrvSprites             = Next; Next += 0x080 * 16 * 16;
	DrvPalette             = (unsigned int*)Next; Next += 576 * sizeof(unsigned int);

	MemEnd                 = Next;

	return 0;
}

static void OpenCPU(int nCPU)
{
	nActiveCPU = nCPU;
	
	switch (nCPU) {
		case 0: {
			Z80SetContext(&Z80_0);
			
			Z80SetIOReadHandler(Z80_0_Config.Z80In);
			Z80SetIOWriteHandler(Z80_0_Config.Z80Out);
			Z80SetProgramReadHandler(Z80_0_Config.Z80Read);
			Z80SetProgramWriteHandler(Z80_0_Config.Z80Write);
			Z80SetCPUOpArgReadHandler(Z80_0_Config.Z80ReadOpArg);
			Z80SetCPUOpReadHandler(Z80_0_Config.Z80ReadOp);
			return;
		}
		
		case 1: {
			Z80SetContext(&Z80_1);
			
			Z80SetIOReadHandler(Z80_1_Config.Z80In);
			Z80SetIOWriteHandler(Z80_1_Config.Z80Out);
			Z80SetProgramReadHandler(Z80_1_Config.Z80Read);
			Z80SetProgramWriteHandler(Z80_1_Config.Z80Write);
			Z80SetCPUOpArgReadHandler(Z80_1_Config.Z80ReadOpArg);
			Z80SetCPUOpReadHandler(Z80_1_Config.Z80ReadOp);
			return;
		}
		
		case 2: {
			Z80SetContext(&Z80_2);
			
			Z80SetIOReadHandler(Z80_2_Config.Z80In);
			Z80SetIOWriteHandler(Z80_2_Config.Z80Out);
			Z80SetProgramReadHandler(Z80_2_Config.Z80Read);
			Z80SetProgramWriteHandler(Z80_2_Config.Z80Write);
			Z80SetCPUOpArgReadHandler(Z80_2_Config.Z80ReadOpArg);
			Z80SetCPUOpReadHandler(Z80_2_Config.Z80ReadOp);
			return;
		}
	}
}

static void CloseCPU()
{
	switch (nActiveCPU) {
		case 0: {
			Z80GetContext(&Z80_0);
			return;
		}
		
		case 1: {
			Z80GetContext(&Z80_1);
			return;
		}
		
		case 2: {
			Z80GetContext(&Z80_2);
			return;
		}
	}
	
	nActiveCPU = -1;
}

static int DrvDoReset()
{
	for (int i = 0; i < 3; i++) {
		OpenCPU(i);
		Z80Reset();
		CloseCPU();
	}
	
	DrvCPU1FireIRQ = 0;
	DrvCPU2FireIRQ = 0;
	DrvCPU3FireIRQ = 0;
	DrvCPU2Halt = 0;
	DrvCPU3Halt = 0;
	DrvFlipScreen = 0;
	for (int i = 0; i < 6; i++) {
		DrvStarControl[i] = 0;
	}
	DrvStarScrollX = 0;
	DrvStarScrollY = 0;
	
	IOChipCustomCommand = 0;
	IOChipCPU1FireIRQ = 0;
	IOChipMode = 0;
	IOChipCredits = 0;
	IOChipCoinPerCredit = 0;
	IOChipCreditPerCoin = 0;
	for (int i = 0; i < 16; i++) {
		IOChipCustom[i] = 0;
	}

	return 0;
}

unsigned char __fastcall GalagaZ80PortRead(unsigned int a)
{
	a &= 0xff;
	
	switch (a) {
		default: {
			bprintf(PRINT_NORMAL, _T("Z80 #%i Port Read 02%x\n"), nActiveCPU, a);
		}
	}	

	return 0;
}

void __fastcall GalagaZ80PortWrite(unsigned int a, unsigned char d)
{
	a &= 0xff;
	
	switch (a) {
		default: {
			bprintf(PRINT_NORMAL, _T("Z80 #%i Port Write %02x, %02%x\n"), nActiveCPU, a, d);
		}
	}
}

unsigned char __fastcall GalagaZ80ProgRead(unsigned int a)
{
	if (a <= 0x3fff) {
		switch (nActiveCPU) {
			case 0: return DrvZ80Rom1[a];
			case 1: return DrvZ80Rom2[a];
			case 2: return DrvZ80Rom3[a];
		}
	}
	
	if (a >= 0x8000 && a <= 0x87ff) return DrvVideoRam[a - 0x8000];
	if (a >= 0x8800 && a <= 0x8bff) return DrvSharedRam1[a - 0x8800];
	if (a >= 0x9000 && a <= 0x93ff) return DrvSharedRam2[a - 0x9000];
	if (a >= 0x9800 && a <= 0x9bff) return DrvSharedRam3[a - 0x9800];
	
	switch (a) {
		case 0x6800:
		case 0x6801:
		case 0x6802:
		case 0x6803:
		case 0x6804:
		case 0x6805:
		case 0x6806:
		case 0x6807: {
			int Offset = a - 0x6800;
			int Bit0 = (DrvDip[2] >> Offset) & 0x01;
			int Bit1 = (DrvDip[1] >> Offset) & 0x01;

			return Bit0 | (Bit1 << 1);
		}
		
		case 0x7000:
		case 0x7001:
		case 0x7002:
		case 0x7003:
		case 0x7004:
		case 0x7005:
		case 0x7006:
		case 0x7007:
		case 0x7008:
		case 0x7009:
		case 0x700a:
		case 0x700b:
		case 0x700c:
		case 0x700d:
		case 0x700e:
		case 0x700f: {
			int Offset = a - 0x7000;
			
			switch (IOChipCustomCommand) {
				case 0x71:
				case 0xb1: {
					if (Offset == 0) {
						if (IOChipMode) {
							return DrvInput[0] | DrvDip[0];
						} else {
							unsigned char In;
							static unsigned char CoinInserted;
							
							In = DrvInput[0] | DrvDip[0];
							if (In != PrevInValue) {
								if (IOChipCoinPerCredit > 0) {
									if ((In & 0x70) != 0x70 && IOChipCredits < 99) {
										CoinInserted++;
										if (CoinInserted >= IOChipCoinPerCredit) {
											IOChipCredits += IOChipCreditPerCoin;
											CoinInserted = 0;
										}
									}
								} else {
									IOChipCredits = 2;
								}
								
								if ((In & 0x04) == 0) {
									if (IOChipCredits >= 1) IOChipCredits--;
								}
							
								if ((In & 0x08) == 0) {
									if (IOChipCredits >= 2) IOChipCredits -= 2;
								}
							}
							
							PrevInValue = In;
							
							return (IOChipCredits / 10) * 16 + IOChipCredits % 10;
						}
					}
					
					if (Offset == 1) return DrvInput[1];
					if (Offset == 2) return DrvInput[2];
				}
			}
			
			return 0xff;
		}
		
		case 0x7100: {
			return IOChipCustomCommand;
		}
		
		default: {
			bprintf(PRINT_NORMAL, _T("Z80 #%i Read %04x\n"), nActiveCPU, a);
		}
	}
	
	return 0;
}

void __fastcall GalagaZ80ProgWrite(unsigned int a, unsigned char d)
{
	if (a <= 0x3fff) return;
	if (a >= 0x8000 && a <= 0x87ff) { DrvVideoRam[a - 0x8000] = d; return; }
	if (a >= 0x8800 && a <= 0x8bff) { DrvSharedRam1[a - 0x8800] = d; return; }
	if (a >= 0x9000 && a <= 0x93ff) { DrvSharedRam2[a - 0x9000] = d; return; }
	if (a >= 0x9800 && a <= 0x9bff) { DrvSharedRam3[a - 0x9800] = d; return; }
	
	if (a >= 0x6800 && a <= 0x681f) { NamcoSoundWrite(a - 0x6800, d); return; }
	
	switch (a) {
		case 0x6820: {
			DrvCPU1FireIRQ = d & 0x01;
			if (!DrvCPU1FireIRQ) {
			
			}
			return;
		}
		
		case 0x6821: {
			DrvCPU2FireIRQ = d & 0x01;
			if (!DrvCPU2FireIRQ) {
			
			}
			return;
		}
		
		case 0x6822: {
			DrvCPU3FireIRQ = !(d & 0x01);
			return;
		}
		
		case 0x6823: {
			if (!(d & 0x01)) {
				int nActive = nActiveCPU;
				CloseCPU();
				OpenCPU(1);
				Z80Reset();
				CloseCPU();
				OpenCPU(2);
				Z80Reset();
				CloseCPU();
				OpenCPU(nActive);
				DrvCPU2Halt = 1;
				DrvCPU3Halt = 1;
				return;
			} else {
				DrvCPU2Halt = 0;
				DrvCPU3Halt = 0;
			}
		}
		
		case 0x6830: {
			// watchdog write
			return;
		}
		
		case 0x7000:
		case 0x7001:
		case 0x7002:
		case 0x7003:
		case 0x7004:
		case 0x7005:
		case 0x7006:
		case 0x7007:
		case 0x7008:
		case 0x7009:
		case 0x700a:
		case 0x700b:
		case 0x700c:
		case 0x700d:
		case 0x700e:
		case 0x700f: {
			int Offset = a - 0x7000;
			IOChipCustom[Offset] = d;
			
			switch (IOChipCustomCommand) {
				case 0xe1: {
					if (Offset == 7) {
						IOChipCoinPerCredit = IOChipCustom[1];
						IOChipCreditPerCoin = IOChipCustom[2];
					}
					break;
				}
			}
			
			return;
		}
	
		case 0x7100: {
			IOChipCustomCommand = d;
			IOChipCPU1FireIRQ = 1;
			
			switch (IOChipCustomCommand) {
				case 0x10: {
					IOChipCPU1FireIRQ = 0;
					return;
				}
				
				case 0xa1: {
					IOChipMode = 1;
					return;
				}
				
				case 0xe1: {
					IOChipCredits = 0;
					IOChipMode = 0;
					return;
				}
			}
			
			return;
		}
		
		case 0xa000:
		case 0xa001:
		case 0xa002:
		case 0xa003:
		case 0xa004:
		case 0xa005:
		case 0xa006: {
			DrvStarControl[a - 0xa000] = d & 0x01;
			return;
		}
		
		case 0xa007: {
			DrvFlipScreen = d & 0x01;
			return;
		}
		
		default: {
			bprintf(PRINT_NORMAL, _T("Z80 #%i Write %04x, %02x\n"), nActiveCPU, a, d);
		}
	}
}

unsigned char __fastcall GalagaZ80OpRead(unsigned int a)
{
	if (a <= 0x3fff) {
		switch (nActiveCPU) {
			case 0: return DrvZ80Rom1[a];
			case 1: return DrvZ80Rom2[a];
			case 2: return DrvZ80Rom3[a];
		}
	}
	
	switch (a) {
		default: {
			bprintf(PRINT_NORMAL, _T("Z80 #%i Op Read %04x\n"), nActiveCPU, a);
		}
	}
	
	return 0;
}

unsigned char __fastcall GalagaZ80OpArgRead(unsigned int a)
{
	if (a <= 0x3fff) {
		switch (nActiveCPU) {
			case 0: return DrvZ80Rom1[a];
			case 1: return DrvZ80Rom2[a];
			case 2: return DrvZ80Rom3[a];
		}
	}
	
	switch (a) {
		default: {
			bprintf(PRINT_NORMAL, _T("Z80 #%i Op Arg Read %04x\n"), nActiveCPU, a);
		}
	}
	
	return 0;
}

static int CharPlaneOffsets[2]   = { 0, 4 };
static int CharXOffsets[8]       = { 64, 65, 66, 67, 0, 1, 2, 3 };
static int CharYOffsets[8]       = { 0, 8, 16, 24, 32, 40, 48, 56 };
static int SpritePlaneOffsets[2] = { 0, 4 };
static int SpriteXOffsets[16]    = { 0, 1, 2, 3, 64, 65, 66, 67, 128, 129, 130, 131, 192, 193, 194, 195 };
static int SpriteYOffsets[16]    = { 0, 8, 16, 24, 32, 40, 48, 56, 256, 264, 272, 280, 288, 296, 304, 312 };

static void MachineInit()
{
	// Setup the Z80 emulation
	Z80Init();
	
	Z80_0_Config.Z80In = GalagaZ80PortRead;
	Z80_0_Config.Z80Out = GalagaZ80PortWrite;
	Z80_0_Config.Z80Read = GalagaZ80ProgRead;
	Z80_0_Config.Z80Write = GalagaZ80ProgWrite;
	Z80_0_Config.Z80ReadOpArg = GalagaZ80OpArgRead;
	Z80_0_Config.Z80ReadOp = GalagaZ80OpRead;
	
	Z80_1_Config.Z80In = GalagaZ80PortRead;
	Z80_1_Config.Z80Out = GalagaZ80PortWrite;
	Z80_1_Config.Z80Read = GalagaZ80ProgRead;
	Z80_1_Config.Z80Write = GalagaZ80ProgWrite;
	Z80_1_Config.Z80ReadOpArg = GalagaZ80OpArgRead;
	Z80_1_Config.Z80ReadOp = GalagaZ80OpRead;
	
	Z80_2_Config.Z80In = GalagaZ80PortRead;
	Z80_2_Config.Z80Out = GalagaZ80PortWrite;
	Z80_2_Config.Z80Read = GalagaZ80ProgRead;
	Z80_2_Config.Z80Write = GalagaZ80ProgWrite;
	Z80_2_Config.Z80ReadOpArg = GalagaZ80OpArgRead;
	Z80_2_Config.Z80ReadOp = GalagaZ80OpRead;
	
	NamcoSoundInit(18432000 / 6 / 32);

	GenericTilesInit();

	// Reset the driver
	DrvDoReset();
}

static int DrvInit()
{
	int nRet = 0, nLen;

	// Allocate and Blank all required memory
	Mem = NULL;
	MemIndex();
	nLen = MemEnd - (unsigned char *)0;
	if ((Mem = (unsigned char *)malloc(nLen)) == NULL) return 1;
	memset(Mem, 0, nLen);
	MemIndex();

	DrvTempRom = (unsigned char *)malloc(0x02000);

	// Load Z80 #1 Program Roms
	nRet = BurnLoadRom(DrvZ80Rom1 + 0x00000,  0, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvZ80Rom1 + 0x01000,  1, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvZ80Rom1 + 0x02000,  2, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvZ80Rom1 + 0x03000,  3, 1); if (nRet != 0) return 1;
	
	// Load Z80 #2 Program Roms
	nRet = BurnLoadRom(DrvZ80Rom2 + 0x00000,  4, 1); if (nRet != 0) return 1;
	
	// Load Z80 #3 Program Roms
	nRet = BurnLoadRom(DrvZ80Rom3 + 0x00000,  5, 1); if (nRet != 0) return 1;
	
	// Load and decode the chars
	nRet = BurnLoadRom(DrvTempRom,            8, 1); if (nRet != 0) return 1;
	GfxDecode(0x100, 2, 8, 8, CharPlaneOffsets, CharXOffsets, CharYOffsets, 0x80, DrvTempRom, DrvChars);
	
	// Load and decode the sprites
	memset(DrvTempRom, 0, 0x02000);
	nRet = BurnLoadRom(DrvTempRom + 0x00000,  9, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvTempRom + 0x01000, 10, 1); if (nRet != 0) return 1;
	GfxDecode(0x80, 2, 16, 16, SpritePlaneOffsets, SpriteXOffsets, SpriteYOffsets, 0x200, DrvTempRom, DrvSprites);

	// Load the PROMs
	nRet = BurnLoadRom(DrvPromPalette,       11, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvPromCharLookup,    12, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvPromSpriteLookup,  13, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(NamcoSoundProm,       14, 1); if (nRet != 0) return 1;
	
	free(DrvTempRom);
	
	MachineInit();

	return 0;
}

static int GallagInit()
{
	int nRet = 0, nLen;

	// Allocate and Blank all required memory
	Mem = NULL;
	MemIndex();
	nLen = MemEnd - (unsigned char *)0;
	if ((Mem = (unsigned char *)malloc(nLen)) == NULL) return 1;
	memset(Mem, 0, nLen);
	MemIndex();

	DrvTempRom = (unsigned char *)malloc(0x02000);

	// Load Z80 #1 Program Roms
	nRet = BurnLoadRom(DrvZ80Rom1 + 0x00000,  0, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvZ80Rom1 + 0x01000,  1, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvZ80Rom1 + 0x02000,  2, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvZ80Rom1 + 0x03000,  3, 1); if (nRet != 0) return 1;
	
	// Load Z80 #2 Program Roms
	nRet = BurnLoadRom(DrvZ80Rom2 + 0x00000,  4, 1); if (nRet != 0) return 1;
	
	// Load Z80 #3 Program Roms
	nRet = BurnLoadRom(DrvZ80Rom3 + 0x00000,  5, 1); if (nRet != 0) return 1;
	
	// Load and decode the chars
	nRet = BurnLoadRom(DrvTempRom,            7, 1); if (nRet != 0) return 1;
	GfxDecode(0x100, 2, 8, 8, CharPlaneOffsets, CharXOffsets, CharYOffsets, 0x80, DrvTempRom, DrvChars);
	
	// Load and decode the sprites
	memset(DrvTempRom, 0, 0x02000);
	nRet = BurnLoadRom(DrvTempRom + 0x00000,  8, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvTempRom + 0x01000,  9, 1); if (nRet != 0) return 1;
	GfxDecode(0x80, 2, 16, 16, SpritePlaneOffsets, SpriteXOffsets, SpriteYOffsets, 0x200, DrvTempRom, DrvSprites);

	// Load the PROMs
	nRet = BurnLoadRom(DrvPromPalette,       10, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvPromCharLookup,    11, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(DrvPromSpriteLookup,  12, 1); if (nRet != 0) return 1;
	nRet = BurnLoadRom(NamcoSoundProm,       13, 1); if (nRet != 0) return 1;
	
	free(DrvTempRom);
	
	MachineInit();

	return 0;
}

static int DrvExit()
{
	GenericTilesExit();
	NamcoSoundExit();
	Z80Exit();
	
	free(Mem);
	Mem = NULL;
	
	DrvCPU1FireIRQ = 0;
	DrvCPU2FireIRQ = 0;
	DrvCPU3FireIRQ = 0;
	DrvCPU2Halt = 0;
	DrvCPU3Halt = 0;
	DrvFlipScreen = 0;
	for (int i = 0; i < 6; i++) {
		DrvStarControl[i] = 0;
	}
	DrvStarScrollX = 0;
	DrvStarScrollY = 0;
	
	IOChipCustomCommand = 0;
	IOChipCPU1FireIRQ = 0;
	IOChipMode = 0;
	IOChipCredits = 0;
	IOChipCoinPerCredit = 0;
	IOChipCreditPerCoin = 0;
	for (int i = 0; i < 16; i++) {
		IOChipCustom[i] = 0;
	}
	
	nActiveCPU = 0;

	return 0;
}

static void DrvCalcPalette()
{
	int i;
	unsigned int Palette[96];
	
	for (i = 0; i < 32; i++) {
		int bit0, bit1, bit2, r, g, b;
		
		bit0 = (DrvPromPalette[i] >> 0) & 0x01;
		bit1 = (DrvPromPalette[i] >> 1) & 0x01;
		bit2 = (DrvPromPalette[i] >> 2) & 0x01;
		r = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;
		bit0 = (DrvPromPalette[i] >> 3) & 0x01;
		bit1 = (DrvPromPalette[i] >> 4) & 0x01;
		bit2 = (DrvPromPalette[i] >> 5) & 0x01;
		g = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;
		bit0 = 0;
		bit1 = (DrvPromPalette[i] >> 6) & 0x01;
		bit2 = (DrvPromPalette[i] >> 7) & 0x01;
		b = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;
		
		Palette[i] = BurnHighCol(r, g, b, 0);
	}
	
	for (i = 0; i < 64; i++) {
		int bits, r, g, b;
		static const int map[4] = { 0x00, 0x47, 0x97, 0xde };
		
		bits = (i >> 0) & 0x03;
		r = map[bits];
		bits = (i >> 2) & 0x03;
		g = map[bits];
		bits = (i >> 4) & 0x03;
		b = map[bits];
		
		Palette[32 + i] = BurnHighCol(r, g, b, 0);
	}
	
	for (i = 0; i < 256; i++) {
		DrvPalette[i] = Palette[((DrvPromCharLookup[i]) & 0x0f) + 0x10];
	}
	
	for (i = 0; i < 256; i++) {
		DrvPalette[256 + i] = Palette[DrvPromSpriteLookup[i] & 0x0f];
	}
	
	for (i = 0; i < 64; i++) {
		DrvPalette[512 + i] = Palette[32 + i];
	}
}

struct Star {
	unsigned short x, y;
	unsigned char Colour, Set;
};

const struct Star StarSeedTab[252] = {
	// set 0
	{0x0085, 0x0006, 0x35, 0x00},
	{0x008F, 0x0008, 0x30, 0x00},
	{0x00E5, 0x001B, 0x07, 0x00},
	{0x0022, 0x001C, 0x31, 0x00},
	{0x00E5, 0x0025, 0x1D, 0x00},
	{0x0015, 0x0026, 0x29, 0x00},
	{0x0080, 0x002D, 0x3B, 0x00},
	{0x0097, 0x002E, 0x1C, 0x00},
	{0x00BA, 0x003B, 0x05, 0x00},
	{0x0036, 0x003D, 0x36, 0x00},
	{0x0057, 0x0044, 0x09, 0x00},
	{0x00CF, 0x0044, 0x3D, 0x00},
	{0x0061, 0x004E, 0x27, 0x00},
	{0x0087, 0x0064, 0x1A, 0x00},
	{0x00D6, 0x0064, 0x17, 0x00},
	{0x000B, 0x006C, 0x3C, 0x00},
	{0x0006, 0x006D, 0x24, 0x00},
	{0x0018, 0x006E, 0x3A, 0x00},
	{0x00A9, 0x0079, 0x23, 0x00},
	{0x008A, 0x007B, 0x11, 0x00},
	{0x00D6, 0x0080, 0x0C, 0x00},
	{0x0067, 0x0082, 0x3F, 0x00},
	{0x0039, 0x0083, 0x38, 0x00},
	{0x0072, 0x0083, 0x14, 0x00},
	{0x00EC, 0x0084, 0x16, 0x00},
	{0x008E, 0x0085, 0x10, 0x00},
	{0x0020, 0x0088, 0x25, 0x00},
	{0x0095, 0x008A, 0x0F, 0x00},
	{0x000E, 0x008D, 0x00, 0x00},
	{0x0006, 0x0091, 0x2E, 0x00},
	{0x0007, 0x0094, 0x0D, 0x00},
	{0x00AE, 0x0097, 0x0B, 0x00},
	{0x0000, 0x0098, 0x2D, 0x00},
	{0x0086, 0x009B, 0x01, 0x00},
	{0x0058, 0x00A1, 0x34, 0x00},
	{0x00FE, 0x00A1, 0x3E, 0x00},
	{0x00A2, 0x00A8, 0x1F, 0x00},
	{0x0041, 0x00AA, 0x0A, 0x00},
	{0x003F, 0x00AC, 0x32, 0x00},
	{0x00DE, 0x00AC, 0x03, 0x00},
	{0x00D4, 0x00B9, 0x26, 0x00},
	{0x006D, 0x00BB, 0x1B, 0x00},
	{0x0062, 0x00BD, 0x39, 0x00},
	{0x00C9, 0x00BE, 0x18, 0x00},
	{0x006C, 0x00C1, 0x04, 0x00},
	{0x0059, 0x00C3, 0x21, 0x00},
	{0x0060, 0x00CC, 0x0E, 0x00},
	{0x0091, 0x00CC, 0x12, 0x00},
	{0x003F, 0x00CF, 0x06, 0x00},
	{0x00F7, 0x00CF, 0x22, 0x00},
	{0x0044, 0x00D0, 0x33, 0x00},
	{0x0034, 0x00D2, 0x08, 0x00},
	{0x00D3, 0x00D9, 0x20, 0x00},
	{0x0071, 0x00DD, 0x37, 0x00},
	{0x0073, 0x00E1, 0x2C, 0x00},
	{0x00B9, 0x00E3, 0x2F, 0x00},
	{0x00A9, 0x00E4, 0x13, 0x00},
	{0x00D3, 0x00E7, 0x19, 0x00},
	{0x0037, 0x00ED, 0x02, 0x00},
	{0x00BD, 0x00F4, 0x15, 0x00},
	{0x000F, 0x00F6, 0x28, 0x00},
	{0x004F, 0x00F7, 0x2B, 0x00},
	{0x00FB, 0x00FF, 0x2A, 0x00},

	// set 1
	{0x00FE, 0x0004, 0x3D, 0x01},
	{0x00C4, 0x0006, 0x10, 0x01},
	{0x001E, 0x0007, 0x2D, 0x01},
	{0x0083, 0x000B, 0x1F, 0x01},
	{0x002E, 0x000D, 0x3C, 0x01},
	{0x001F, 0x000E, 0x00, 0x01},
	{0x00D8, 0x000E, 0x2C, 0x01},
	{0x0003, 0x000F, 0x17, 0x01},
	{0x0095, 0x0011, 0x3F, 0x01},
	{0x006A, 0x0017, 0x35, 0x01},
	{0x00CC, 0x0017, 0x02, 0x01},
	{0x0000, 0x0018, 0x32, 0x01},
	{0x0092, 0x001D, 0x36, 0x01},
	{0x00E3, 0x0021, 0x04, 0x01},
	{0x002F, 0x002D, 0x37, 0x01},
	{0x00F0, 0x002F, 0x0C, 0x01},
	{0x009B, 0x003E, 0x06, 0x01},
	{0x00A4, 0x004C, 0x07, 0x01},
	{0x00EA, 0x004D, 0x13, 0x01},
	{0x0084, 0x004E, 0x21, 0x01},
	{0x0033, 0x0052, 0x0F, 0x01},
	{0x0070, 0x0053, 0x0E, 0x01},
	{0x0006, 0x0059, 0x08, 0x01},
	{0x0081, 0x0060, 0x28, 0x01},
	{0x0037, 0x0061, 0x29, 0x01},
	{0x008F, 0x0067, 0x2F, 0x01},
	{0x001B, 0x006A, 0x1D, 0x01},
	{0x00BF, 0x007C, 0x12, 0x01},
	{0x0051, 0x007F, 0x31, 0x01},
	{0x0061, 0x0086, 0x25, 0x01},
	{0x006A, 0x008F, 0x0D, 0x01},
	{0x006A, 0x0091, 0x19, 0x01},
	{0x0090, 0x0092, 0x05, 0x01},
	{0x003B, 0x0096, 0x24, 0x01},
	{0x008C, 0x0097, 0x0A, 0x01},
	{0x0006, 0x0099, 0x03, 0x01},
	{0x0038, 0x0099, 0x38, 0x01},
	{0x00A8, 0x0099, 0x18, 0x01},
	{0x0076, 0x00A6, 0x20, 0x01},
	{0x00AD, 0x00A6, 0x1C, 0x01},
	{0x00EC, 0x00A6, 0x1E, 0x01},
	{0x0086, 0x00AC, 0x15, 0x01},
	{0x0078, 0x00AF, 0x3E, 0x01},
	{0x007B, 0x00B3, 0x09, 0x01},
	{0x0027, 0x00B8, 0x39, 0x01},
	{0x0088, 0x00C2, 0x23, 0x01},
	{0x0044, 0x00C3, 0x3A, 0x01},
	{0x00CF, 0x00C5, 0x34, 0x01},
	{0x0035, 0x00C9, 0x30, 0x01},
	{0x006E, 0x00D1, 0x3B, 0x01},
	{0x00D6, 0x00D7, 0x16, 0x01},
	{0x003A, 0x00D9, 0x2B, 0x01},
	{0x00AB, 0x00E0, 0x11, 0x01},
	{0x00E0, 0x00E2, 0x1B, 0x01},
	{0x006F, 0x00E6, 0x0B, 0x01},
	{0x00B8, 0x00E8, 0x14, 0x01},
	{0x00D9, 0x00E8, 0x1A, 0x01},
	{0x00F9, 0x00E8, 0x22, 0x01},
	{0x0004, 0x00F1, 0x2E, 0x01},
	{0x0049, 0x00F8, 0x26, 0x01},
	{0x0010, 0x00F9, 0x01, 0x01},
	{0x0039, 0x00FB, 0x33, 0x01},
	{0x0028, 0x00FC, 0x27, 0x01},

	// set 2
	{0x00FA, 0x0006, 0x19, 0x02},
	{0x00E4, 0x0007, 0x2D, 0x02},
	{0x0072, 0x000A, 0x03, 0x02},
	{0x0084, 0x001B, 0x00, 0x02},
	{0x00BA, 0x001D, 0x29, 0x02},
	{0x00E3, 0x0022, 0x04, 0x02},
	{0x00D1, 0x0026, 0x2A, 0x02},
	{0x0089, 0x0032, 0x30, 0x02},
	{0x005B, 0x0036, 0x27, 0x02},
	{0x0084, 0x003A, 0x36, 0x02},
	{0x0053, 0x003F, 0x0D, 0x02},
	{0x0008, 0x0040, 0x1D, 0x02},
	{0x0055, 0x0040, 0x1A, 0x02},
	{0x00AA, 0x0041, 0x31, 0x02},
	{0x00FB, 0x0041, 0x2B, 0x02},
	{0x00BC, 0x0046, 0x16, 0x02},
	{0x0093, 0x0052, 0x39, 0x02},
	{0x00B9, 0x0057, 0x10, 0x02},
	{0x0054, 0x0059, 0x28, 0x02},
	{0x00E6, 0x005A, 0x01, 0x02},
	{0x00A7, 0x005D, 0x1B, 0x02},
	{0x002D, 0x005E, 0x35, 0x02},
	{0x0014, 0x0062, 0x21, 0x02},
	{0x0069, 0x006D, 0x1F, 0x02},
	{0x00CE, 0x006F, 0x0B, 0x02},
	{0x00DF, 0x0075, 0x2F, 0x02},
	{0x00CB, 0x0077, 0x12, 0x02},
	{0x004E, 0x007C, 0x23, 0x02},
	{0x004A, 0x0084, 0x0F, 0x02},
	{0x0012, 0x0086, 0x25, 0x02},
	{0x0068, 0x008C, 0x32, 0x02},
	{0x0003, 0x0095, 0x20, 0x02},
	{0x000A, 0x009C, 0x17, 0x02},
	{0x005B, 0x00A3, 0x08, 0x02},
	{0x005F, 0x00A4, 0x3E, 0x02},
	{0x0072, 0x00A4, 0x2E, 0x02},
	{0x00CC, 0x00A6, 0x06, 0x02},
	{0x008A, 0x00AB, 0x0C, 0x02},
	{0x00E0, 0x00AD, 0x26, 0x02},
	{0x00F3, 0x00AF, 0x0A, 0x02},
	{0x0075, 0x00B4, 0x13, 0x02},
	{0x0068, 0x00B7, 0x11, 0x02},
	{0x006D, 0x00C2, 0x2C, 0x02},
	{0x0076, 0x00C3, 0x14, 0x02},
	{0x00CF, 0x00C4, 0x1E, 0x02},
	{0x0004, 0x00C5, 0x1C, 0x02},
	{0x0013, 0x00C6, 0x3F, 0x02},
	{0x00B9, 0x00C7, 0x3C, 0x02},
	{0x0005, 0x00D7, 0x34, 0x02},
	{0x0095, 0x00D7, 0x3A, 0x02},
	{0x00FC, 0x00D8, 0x02, 0x02},
	{0x00E7, 0x00DC, 0x09, 0x02},
	{0x001D, 0x00E1, 0x05, 0x02},
	{0x0005, 0x00E6, 0x33, 0x02},
	{0x001C, 0x00E9, 0x3B, 0x02},
	{0x00A2, 0x00ED, 0x37, 0x02},
	{0x0028, 0x00EE, 0x07, 0x02},
	{0x00DD, 0x00EF, 0x18, 0x02},
	{0x006D, 0x00F0, 0x38, 0x02},
	{0x00A1, 0x00F2, 0x0E, 0x02},
	{0x0074, 0x00F7, 0x3D, 0x02},
	{0x0069, 0x00F9, 0x22, 0x02},
	{0x003F, 0x00FF, 0x24, 0x02},

	// set 3
	{0x0071, 0x0010, 0x34, 0x03},
	{0x00AF, 0x0011, 0x23, 0x03},
	{0x00A0, 0x0014, 0x26, 0x03},
	{0x0002, 0x0017, 0x02, 0x03},
	{0x004B, 0x0019, 0x31, 0x03},
	{0x0093, 0x001C, 0x0E, 0x03},
	{0x001B, 0x001E, 0x25, 0x03},
	{0x0032, 0x0020, 0x2E, 0x03},
	{0x00EE, 0x0020, 0x3A, 0x03},
	{0x0079, 0x0022, 0x2F, 0x03},
	{0x006C, 0x0023, 0x17, 0x03},
	{0x00BC, 0x0025, 0x11, 0x03},
	{0x0041, 0x0029, 0x30, 0x03},
	{0x001C, 0x002E, 0x32, 0x03},
	{0x00B9, 0x0031, 0x01, 0x03},
	{0x0083, 0x0032, 0x05, 0x03},
	{0x0095, 0x003A, 0x12, 0x03},
	{0x000D, 0x003F, 0x07, 0x03},
	{0x0020, 0x0041, 0x33, 0x03},
	{0x0092, 0x0045, 0x2C, 0x03},
	{0x00D4, 0x0047, 0x08, 0x03},
	{0x00A1, 0x004B, 0x2D, 0x03},
	{0x00D2, 0x004B, 0x3B, 0x03},
	{0x00D6, 0x0052, 0x24, 0x03},
	{0x009A, 0x005F, 0x1C, 0x03},
	{0x0016, 0x0060, 0x3D, 0x03},
	{0x001A, 0x0063, 0x1F, 0x03},
	{0x00CD, 0x0066, 0x28, 0x03},
	{0x00FF, 0x0067, 0x10, 0x03},
	{0x0035, 0x0069, 0x20, 0x03},
	{0x008F, 0x006C, 0x04, 0x03},
	{0x00CA, 0x006C, 0x2A, 0x03},
	{0x005A, 0x0074, 0x09, 0x03},
	{0x0060, 0x0078, 0x38, 0x03},
	{0x0072, 0x0079, 0x1E, 0x03},
	{0x0037, 0x007F, 0x29, 0x03},
	{0x0012, 0x0080, 0x14, 0x03},
	{0x0029, 0x0082, 0x2B, 0x03},
	{0x0084, 0x0098, 0x36, 0x03},
	{0x0032, 0x0099, 0x37, 0x03},
	{0x00BB, 0x00A0, 0x19, 0x03},
	{0x003E, 0x00A3, 0x3E, 0x03},
	{0x004A, 0x00A6, 0x1A, 0x03},
	{0x0029, 0x00A7, 0x21, 0x03},
	{0x009D, 0x00B7, 0x22, 0x03},
	{0x006C, 0x00B9, 0x15, 0x03},
	{0x000C, 0x00C0, 0x0A, 0x03},
	{0x00C2, 0x00C3, 0x0F, 0x03},
	{0x002F, 0x00C9, 0x0D, 0x03},
	{0x00D2, 0x00CE, 0x16, 0x03},
	{0x00F3, 0x00CE, 0x0B, 0x03},
	{0x0075, 0x00CF, 0x27, 0x03},
	{0x001A, 0x00D5, 0x35, 0x03},
	{0x0026, 0x00D6, 0x39, 0x03},
	{0x0080, 0x00DA, 0x3C, 0x03},
	{0x00A9, 0x00DD, 0x00, 0x03},
	{0x00BC, 0x00EB, 0x03, 0x03},
	{0x0032, 0x00EF, 0x1B, 0x03},
	{0x0067, 0x00F0, 0x3F, 0x03},
	{0x00EF, 0x00F1, 0x18, 0x03},
	{0x00A8, 0x00F3, 0x0C, 0x03},
	{0x00DE, 0x00F9, 0x1D, 0x03},
	{0x002C, 0x00FA, 0x13, 0x03}
};

static void DrvRenderStars()
{
	if (DrvStarControl[5] == 1) {
  		int StarCounter;
		int SetA, SetB;

		SetA = DrvStarControl[3];
		SetB = DrvStarControl[4] | 0x02;

		for (StarCounter = 0; StarCounter < 252; StarCounter++) {
			int x, y;

			if ((SetA == StarSeedTab[StarCounter].Set) || (SetB == StarSeedTab[StarCounter].Set)) {
				x = (StarSeedTab[StarCounter].x + DrvStarScrollX) % 256 + 16;
				y = (112 + StarSeedTab[StarCounter].y + DrvStarScrollY) % 256;

				if (x >= 0 && x < 288 && y >= 0 && y < 224) {
					pTransDraw[(y * nScreenWidth) + x] = StarSeedTab[StarCounter].Colour + 512;
				}
			}

		}
	}
}

static void DrvRenderTilemap()
{
	int mx, my, Code, Colour, x, y, TileIndex, Row, Col;

	for (mx = 0; mx < 28; mx++) {
		for (my = 0; my < 36; my++) {
			Row = mx + 2;
			Col = my - 2;
			if (Col & 0x20) {
				TileIndex = Row + ((Col & 0x1f) << 5);
			} else {
				TileIndex = Col + (Row << 5);
			}
			
			Code = DrvVideoRam[TileIndex + 0x000] & 0x7f;
			Colour = DrvVideoRam[TileIndex + 0x400] & 0x3f;

			y = 8 * mx;
			x = 8 * my;
			
			if (DrvFlipScreen) {
				x = 280 - x;
				y = 216 - y;
			}
			
			if (x > 8 && x < 280 && y > 8 && y < 216) {
				if (DrvFlipScreen) {
					Render8x8Tile_FlipXY(pTransDraw, Code, x, y, Colour, 2, 0, DrvChars);
				} else {
					Render8x8Tile(pTransDraw, Code, x, y, Colour, 2, 0, DrvChars);
				}
			} else {
				if (DrvFlipScreen) {
					Render8x8Tile_FlipXY_Clip(pTransDraw, Code, x, y, Colour, 2, 0, DrvChars);
				} else {
					Render8x8Tile_Clip(pTransDraw, Code, x, y, Colour, 2, 0, DrvChars);
				}
			}
		}
	}
}

static void DrvRenderSprites()
{
	unsigned char *SpriteRam1 = DrvSharedRam1 + 0x380;
	unsigned char *SpriteRam2 = DrvSharedRam2 + 0x380;
	unsigned char *SpriteRam3 = DrvSharedRam3 + 0x380;
	
	for (int Offset = 0; Offset < 0x80; Offset += 2) {
		static const int GfxOffset[2][2] = {
			{ 0, 1 },
			{ 2, 3 }
		};
		int Sprite = SpriteRam1[Offset + 0] & 0x7f;
		int Colour = SpriteRam1[Offset + 1] & 0x3f;
		int sx = SpriteRam2[Offset + 1] - 40 + (0x100 * (SpriteRam3[Offset + 1] & 0x03));
		int sy = 256 - SpriteRam2[Offset + 0] + 1;
		int xFlip = (SpriteRam3[Offset + 0] & 0x01);
		int yFlip = (SpriteRam3[Offset + 0] & 0x02) >> 1;
		int xSize = (SpriteRam3[Offset + 0] & 0x04) >> 2;
		int ySize = (SpriteRam3[Offset + 0] & 0x08) >> 3;

		sy -= 16 * ySize;
		sy = (sy & 0xff) - 32;

		if (DrvFlipScreen) {
			xFlip = !xFlip;
			yFlip = !yFlip;
		}

		for (int y = 0; y <= ySize; y++) {
			for (int x = 0; x <= xSize; x++) {
				int Code = Sprite + GfxOffset[y ^ (ySize * yFlip)][x ^ (xSize * xFlip)];
				int xPos = sx + 16 * x;
				int yPos = sy + 16 * y;
				
				if (xPos > 16 && xPos < 272 && yPos > 16 && yPos < 208) {
					if (xFlip) {
						if (yFlip) {
							Render16x16Tile_Mask_FlipXY(pTransDraw, Code, xPos, yPos, Colour, 2, 0, 256, DrvSprites);
						} else {
							Render16x16Tile_Mask_FlipX(pTransDraw, Code, xPos, yPos, Colour, 2, 0, 256, DrvSprites);
						}
					} else {
						if (yFlip) {
							Render16x16Tile_Mask_FlipY(pTransDraw, Code, xPos, yPos, Colour, 2, 0, 256, DrvSprites);
						} else {
							Render16x16Tile_Mask(pTransDraw, Code, xPos, yPos, Colour, 2, 0, 256, DrvSprites);
						}
					}
				} else {
					if (xFlip) {
						if (yFlip) {
							Render16x16Tile_Mask_FlipXY_Clip(pTransDraw, Code, xPos, yPos, Colour, 2, 0, 256, DrvSprites);
						} else {
							Render16x16Tile_Mask_FlipX_Clip(pTransDraw, Code, xPos, yPos, Colour, 2, 0, 256, DrvSprites);
						}
					} else {
						if (yFlip) {
							Render16x16Tile_Mask_FlipY_Clip(pTransDraw, Code, xPos, yPos, Colour, 2, 0, 256, DrvSprites);
						} else {
							Render16x16Tile_Mask_Clip(pTransDraw, Code, xPos, yPos, Colour, 2, 0, 256, DrvSprites);
						}
					}
				}
			}
		}
	}
}

static void DrvDraw()
{
	BurnTransferClear();
	DrvCalcPalette();
	DrvRenderTilemap();
	DrvRenderStars();
	DrvRenderSprites();	
	BurnTransferCopy(DrvPalette);
}

static int DrvFrame()
{
	int nInterleave = nBurnSoundLen;
	
	if (DrvReset) DrvDoReset();

	DrvMakeInputs();
	
	int nSoundBufferPos = 0;

	nCyclesTotal[0] = (18432000 / 6) / 60;
	nCyclesTotal[1] = (18432000 / 6) / 60;
	nCyclesTotal[2] = (18432000 / 6) / 60;
	nCyclesDone[0] = nCyclesDone[1] = nCyclesDone[2] = 0;
	
	for (int i = 0; i < nInterleave; i++) {
		int nCurrentCPU, nNext;
		
		// Run Z80 #1
		nCurrentCPU = 0;
		OpenCPU(nCurrentCPU);
		nNext = (i + 1) * nCyclesTotal[nCurrentCPU] / nInterleave;
		nCyclesSegment = nNext - nCyclesDone[nCurrentCPU];
		nCyclesSegment = Z80Execute(nCyclesSegment);
		nCyclesDone[nCurrentCPU] += nCyclesSegment;
		if (i == (nInterleave - 1) && DrvCPU1FireIRQ) {
			Z80SetIrqLine(0, 1);
			Z80Execute(0);
			Z80SetIrqLine(0, 0);
			Z80Execute(0);
		}		
		if ((i == (nInterleave / 4) || i == (nInterleave / 4 * 2) || i == (nInterleave / 4 * 3)) && IOChipCPU1FireIRQ) {
			Z80SetIrqLine(Z80_INPUT_LINE_NMI, 1);
			Z80Execute(0);
			Z80SetIrqLine(Z80_INPUT_LINE_NMI, 0);
			Z80Execute(0);	
		}
		CloseCPU();
		
		if (!DrvCPU2Halt) {
			nCurrentCPU = 1;
			OpenCPU(nCurrentCPU);
			nNext = (i + 1) * nCyclesTotal[nCurrentCPU] / nInterleave;
			nCyclesSegment = nNext - nCyclesDone[nCurrentCPU];
			nCyclesSegment = Z80Execute(nCyclesSegment);
			nCyclesDone[nCurrentCPU] += nCyclesSegment;
			if (i == (nInterleave - 1) && DrvCPU2FireIRQ) {
				Z80SetIrqLine(0, 1);
				Z80Execute(0);
				Z80SetIrqLine(0, 0);
				Z80Execute(0);
			}
			CloseCPU();
		}
		
		if (!DrvCPU3Halt) {
			nCurrentCPU = 2;
			OpenCPU(nCurrentCPU);
			nNext = (i + 1) * nCyclesTotal[nCurrentCPU] / nInterleave;
			nCyclesSegment = nNext - nCyclesDone[nCurrentCPU];
			nCyclesSegment = Z80Execute(nCyclesSegment);
			nCyclesDone[nCurrentCPU] += nCyclesSegment;
			if ((i == (nInterleave / 2) || i == (nInterleave - 1)) && DrvCPU3FireIRQ) {
				Z80SetIrqLine(Z80_INPUT_LINE_NMI, 1);
				Z80Execute(0);
				Z80SetIrqLine(Z80_INPUT_LINE_NMI, 0);
				Z80Execute(0);
			}
			CloseCPU();
		}
		
		if (pBurnSoundOut) {
			int nSegmentLength = nBurnSoundLen / nInterleave;
			short* pSoundBuf = pBurnSoundOut + (nSoundBufferPos << 1);
			
			if (nSegmentLength) {
				NamcoSoundUpdate(pSoundBuf, nSegmentLength);
			}
			nSoundBufferPos += nSegmentLength;
		}
	}
	
	if (pBurnSoundOut) {
		int nSegmentLength = nBurnSoundLen - nSoundBufferPos;
		short* pSoundBuf = pBurnSoundOut + (nSoundBufferPos << 1);

		if (nSegmentLength) {
			NamcoSoundUpdate(pSoundBuf, nSegmentLength);
		}
	}

	if (pBurnDraw) DrvDraw();
	
	static const int Speeds[8] = { -1, -2, -3, 0, 3, 2, 1, 0 };

	DrvStarScrollX += Speeds[DrvStarControl[0] + (DrvStarControl[1] * 2) + (DrvStarControl[2] * 4)];

	return 0;
}

static int DrvScan(int nAction, int *pnMin)
{
	struct BurnArea ba;
	
	if (pnMin != NULL) {			// Return minimum compatible version
		*pnMin = 0x029698;
	}

	if (nAction & ACB_MEMORY_RAM) {
		memset(&ba, 0, sizeof(ba));
		ba.Data	  = RamStart;
		ba.nLen	  = RamEnd-RamStart;
		ba.szName = "All Ram";
		BurnAcb(&ba);
	}
	return 0;
}

struct BurnDriver BurnDrvGalaga = {
	"galaga", NULL, NULL, "1981",
	"Galaga (Namco rev. B)\0", NULL, "Namco", "Miscellaneous",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_ORIENTATION_VERTICAL | BDF_ORIENTATION_FLIPPED, 2, HARDWARE_MISC_PRE90S, GBF_VERSHOOT, 0,
	NULL, DrvRomInfo, DrvRomName, DrvInputInfo, DrvDIPInfo,
	DrvInit, DrvExit, DrvFrame, NULL, DrvScan,
	0, NULL, NULL, NULL, NULL, 224, 288, 3, 4
};

struct BurnDriver BurnDrvGalagao = {
	"galagao", "galaga", NULL, "1981",
	"Galaga (Namco)\0", NULL, "Namco", "Miscellaneous",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL | BDF_ORIENTATION_FLIPPED, 2, HARDWARE_MISC_PRE90S, GBF_VERSHOOT, 0,
	NULL, GalagaoRomInfo, GalagaoRomName, DrvInputInfo, DrvDIPInfo,
	DrvInit, DrvExit, DrvFrame, NULL, DrvScan,
	0, NULL, NULL, NULL, NULL, 224, 288, 3, 4
};

struct BurnDriver BurnDrvGalagamw = {
	"galagamw", "galaga", NULL, "1981",
	"Galaga (Midway set 1)\0", NULL, "Namco (Midway License)", "Miscellaneous",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL | BDF_ORIENTATION_FLIPPED, 2, HARDWARE_MISC_PRE90S, GBF_VERSHOOT, 0,
	NULL, GalagamwRomInfo, GalagamwRomName, DrvInputInfo, GalagamwDIPInfo,
	DrvInit, DrvExit, DrvFrame, NULL, DrvScan,
	0, NULL, NULL, NULL, NULL, 224, 288, 3, 4
};

struct BurnDriver BurnDrvGalagamk = {
	"galagamk", "galaga", NULL, "1981",
	"Galaga (Midway set 2)\0", NULL, "Namco (Midway License)", "Miscellaneous",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL | BDF_ORIENTATION_FLIPPED, 2, HARDWARE_MISC_PRE90S, GBF_VERSHOOT, 0,
	NULL, GalagamkRomInfo, GalagamkRomName, DrvInputInfo, DrvDIPInfo,
	DrvInit, DrvExit, DrvFrame, NULL, DrvScan,
	0, NULL, NULL, NULL, NULL, 224, 288, 3, 4
};

struct BurnDriver BurnDrvGalagamf = {
	"galagamf", "galaga", NULL, "1981",
	"Galaga (Midway set 1 with fast shoot hack)\0", NULL, "Namco (Midway License)", "Miscellaneous",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL | BDF_ORIENTATION_FLIPPED, 2, HARDWARE_MISC_PRE90S, GBF_VERSHOOT, 0,
	NULL, GalagamfRomInfo, GalagamfRomName, DrvInputInfo, DrvDIPInfo,
	DrvInit, DrvExit, DrvFrame, NULL, DrvScan,
	0, NULL, NULL, NULL, NULL, 224, 288, 3, 4
};

struct BurnDriver BurnDrvGallag = {
	"gallag", "galaga", NULL, "1981",
	"Gallag\0", NULL, "bootleg", "Miscellaneous",
	NULL, NULL, NULL, NULL,
	BDF_GAME_WORKING | BDF_CLONE | BDF_ORIENTATION_VERTICAL | BDF_ORIENTATION_FLIPPED | BDF_BOOTLEG, 2, HARDWARE_MISC_PRE90S, GBF_VERSHOOT, 0,
	NULL, GallagRomInfo, GallagRomName, DrvInputInfo, DrvDIPInfo,
	GallagInit, DrvExit, DrvFrame, NULL, DrvScan,
	0, NULL, NULL, NULL, NULL, 224, 288, 3, 4
};
