#include "MessageManager.h"
#include <time.h>
#include <assert.h>
using namespace std;



#ifdef __MINGW32__
  //#define DPRINT(format, ...) if(1)bprintf(0, format, __VA_ARGS__ )
	//#define DPRINT(format, args...) if(1)bprintf(0, ## args)
	#define DPRINT if(FBA_DEBUG)bprintf
  extern int (__cdecl *bprintf) (int nStatus, TCHAR* szFormat, ...);
#else
  #define DPRINT(format, ...)
#endif



#define DATA_MSGS 3 //msgs needed to correctly transmit data: data, ack, ok
#define N_ESTIMATIONS 8
#define N_STRESS_LOAD 1000 //per player
/* ping types*/
#define TIME_PING 0
#define STRESS_PING 1


clock_t lastGreenLight=-1;
/*
 * About byte packing:
 * 640890 = 9C77A
 * buf[0] = 7A
 * buf[1] = C7
 * buf[2] = 09
 *  
 *
 **/
void packInBytes(int toPack, int nBytes, char *dest){
	memcpy(dest, &toPack, nBytes);
}
int getPackedInBytes(char *buf, int nBytes){
  int out = 0;
	memcpy(&out, buf, nBytes);
  return out;
}

int clk2ms(clock_t in){
	return (in*1000)/CLOCKS_PER_SEC ;
}

/*used for inputs to be confirmed (sync, playerNo)*/
int makeSyncAndPlayerNoKey(int frameNo, int playerNo){
  return frameNo*100 + playerNo;
}
int getFrameNo(int key){
  return key/100;
}
int getPlayerNo(int key){
  return key%100;
}


void MessageManager::receiveLoop(){
  outSockSem->enterCritical();
  myMMInfo->looping = true;
  outSockSem->exitCritical();     

  while(myMMInfo->looping){
    tryAndReceive(true);
	}
	DPRINT(0,"MM:Finished receivelooping\n");
}

void MessageManager::retrialLoop(){
	outSockSem->enterCritical();
  myMMInfo->looping = true;
  outSockSem->exitCritical();     

  while(myMMInfo->looping){
    retrial();
	}
	DPRINT(0,"MM:Finished retrialLooping\n");
}

MessageManager::MessageManager(CalleeBack *callback){
	calleeInstance = callback;
	thisGI = callback->getGI();
	outSockSem = new P2pSem();
	myMMInfo = (struct MMInfo *)malloc( sizeof(struct MMInfo) );
  memset(myMMInfo, 0, sizeof(struct MMInfo));
  myMMInfo->maxDelay = -1;
  myMMInfo->loWeight = 6;
	myMMInfo->hiWeight = 1;
  myMMInfo->looping = false;
	myMMInfo->sentMsgCount = 0;
	myMMInfo->allConnected = false;
	
	//headerSize = iData;
	dataSize = -1;
  
	DPRINT(0,"enums:\
	iMsgCount,%d\n\
	iMsgType,%d\n\
	iGameID,%d\n\
	iPingType,%d\n\
	iFrameNo,%d\n\
	iPlayerNo,%d\n\
	iDataSize,%d\n\
	headerSize,%d\n\
	iData,%d\n",
	iMsgCount,
	iMsgType,
	iGameID,
	iPingType,
	iFrameNo,
	iPlayerNo,
	iDataSize,
	headerSize,
	iData);
	
}

void MessageManager::connectUp(queue<string> IPsToConnectTo,unsigned int targetConns){

  DPRINT(0,"Connecting people\n");

  //while(lAddr.size() < targetConns){
	while(myNI.size() < targetConns){
    tryAndReceive(false);
    if(IPsToConnectTo.size() > 0){
      string addr = IPsToConnectTo.front();
      IPsToConnectTo.pop();
      if( !isStored(addr) ){
        invite(addr);
        IPsToConnectTo.push(addr); //go to the back of queue, we'll see if you connected
      }
    }
  }
	
	/*preserve this order: delay, stress*/
	getMaxDelay(); 
  
	string home("home");
  if(bossIP == home){
		stressAll();
	}


}







bool MessageManager::invite(string addr){
  
  
  DPRINT(0,"Inviting %s\n", addr.c_str());

  char buf[headerSize];
	
  memset( buf, 0, sizeof(buf));


  buf[iMsgType] = mtOpenConn;
  
  basicSend(buf, headerSize, addr);
	
  return true;
}

void MessageManager::greet(){
	char buf[headerSize + 2*INT_DIG];
	for(  map<string, P2pNodeInfo*>::iterator lIter = myNI.begin();
				lIter != myNI.end();
				lIter++){

		string currAdr = lIter->first;
		
		memset( buf, 0, sizeof(buf));
		
		buf[iMsgType] = mtPing;
		buf[iPingType] = TIME_PING;
		sprintf(buf+iData,"%d", myMMInfo->maxDelay);
		
		int msTime = clk2ms(clock() - myNI[currAdr]->lastPingTime); 
		
		if( msTime > myNI[currAdr]->greetSpace ){
			myNI[currAdr]->lastPingTime = clock();
			sprintf(buf+iData + INT_DIG,"%u",myNI[currAdr]->lastPingTime);
			DPRINT(0,"Pinging mr. %s having maxDelay of %d (lastPingTime of %u)\n",
						currAdr.c_str(),
						myMMInfo->maxDelay,
						myNI[currAdr]->lastPingTime);
		 basicSend(buf, headerSize + 2*INT_DIG, currAdr);
		}
  }

}



bool MessageManager::isStored(string strin){
  map< string, P2pNodeInfo *>::iterator result = myNI.find(strin);
	return result != myNI.end();
	
}
/*
void MessageManager::store(string addrIn){
	DPRINT(0,"If using WINMM, YOU'RE NOT SEEING THIS!");
  myNI[addrIn] = new P2pNodeInfo();
  
}
*/

void MessageManager::remove(string in){
  map<string, P2pNodeInfo *>::iterator i = myNI.find(in);
  if(i != myNI.end()){
    P2pNodeInfo *hisNI = i->second;
    myNI.erase(i);
    delete hisNI;
  } else
		DPRINT(0,"You shouldn't erase what doesn't exist\n");

  
}





void MessageManager::broadcastData( unsigned char *msg,int msgSize, int frameNoIn){
	if(dataSize < 0)
		dataSize = msgSize;
	
  char *buf = new char[headerSize + dataSize];
  memset(buf, 0, sizeof(buf));
 
 //header
  buf[iMsgType] = mtData;
  buf[iGameID] = thisGI->gameID;
  
	packInBytes(frameNoIn, B_FRAME_NO, buf + iFrameNo);
  
	buf[iPlayerNo] = thisGI->playerNo;
  buf[iDataSize] = msgSize;
	
  //data
  memcpy(buf+iData, msg, msgSize);

 

	
  
  
  //create message
  //when is this cleared?
  //outside critical section because no one will access this key
	
  int key = makeSyncAndPlayerNoKey( frameNoIn, thisGI->playerNo);
  msgInProcess[key] = new P2pMessage( msg, msgSize, thisGI->totalPlayers, frameNoIn);
  msgInProcess[key]->playerConfirm( thisGI->playerNo );
	DPRINT(0,"Created message on key:%d, size:%d and num2conf:%d\n",key ,msgSize, thisGI->totalPlayers);
  //let this broadcast be as atomic as it can... hence the sem is outside the for
	clock_t timer = clock();
	DPRINT(0,"%d:Entering critical\n", clk2ms(clock() - timer));
  outSockSem->enterCritical();
	DPRINT(0,"%d:Entered critical\n", clk2ms(clock() - timer));

  for(  map< string, P2pNodeInfo *>::iterator lIterAddr = myNI.begin();
          lIterAddr != myNI.end();
          lIterAddr++){
		string out = lIterAddr->first;
    basicSend(buf, headerSize + dataSize, out, false);//non-blocking
		lIterAddr->second->lastSentTime = clock();
  }
	DPRINT(0,"%d:Exiting critical\n", clk2ms(clock() - timer));
  outSockSem->exitCritical();
  DPRINT(0,"%d:Exited critical\n", clk2ms(clock() - timer));
	//DPRINT(0,"...broadcast done\n");
	delete[] buf;
}

void MessageManager::broadcastStateLoad( int frameSaved, int frameLoading, int playerNo){
	char *buf = new char[headerSize + 2];
	memset(buf, 0, sizeof(buf));
	
	//header
	buf[iMsgType] = mtLoadState;
	buf[iGameID] = thisGI->gameID;
	
	packInBytes(frameSaved, B_FRAME_NO, buf + iFrameNo);
	buf[iPlayerNo] = thisGI->playerNo;
	
	//data
	buf[iData] = playerNo;
	packInBytes(frameLoading, B_FRAME_NO, buf + iData+1);
	
	
	
	
	outSockSem->enterCritical();
	for(  map< string, P2pNodeInfo *>::iterator lIterAddr = myNI.begin();
          lIterAddr != myNI.end();
          lIterAddr++){
    string out = lIterAddr->first;
    basicSend(buf, headerSize + 2, out, false);//non-blocking
	
	}
	
	outSockSem->exitCritical();
	delete[] buf;
}


void MessageManager::setCallbacks(CalleeBack *callee ){
  calleeInstance = callee;
  
  DPRINT(0,"New callback\n");

}


MessageManager::~MessageManager()
{
	
	looping = false;
	delete outSockSem;
	for( map<string, P2pNodeInfo *>::iterator iAddr = myNI.begin();
			iAddr != myNI.end();
			iAddr++)
		delete iAddr->second;
	myNI.clear();
	ackedFromStress.clear();
	free(myMMInfo);
  
}

void MessageManager::receivedOpenConnection(string addrIn){
  char szBuffer[headerSize];
  memset( szBuffer, 0, sizeof(szBuffer) );
  szBuffer[iMsgType] = mtHandShake;

  basicSend(szBuffer, headerSize, addrIn);
  
  store(addrIn);
  DPRINT(0,"added mr. %s\n",addrIn.c_str() );
	
	
}

void MessageManager::receivedHandShake(string addrIn){

  if(!isStored(addrIn)){
    store(addrIn);
    DPRINT(0,"\nOh, didn't had him... no worry, I've added him\n");
  }
}


void MessageManager::signalReady(struct gameInfo *giIn){
	if(dataSize <0)
		dataSize = thisGI->inputSize;
	
  if( (clock() - lastGreenLight) > (clock_t)1*CLOCKS_PER_SEC){
    DPRINT(0,"I (p%d) am ready!@%u\n", giIn->playerNo,clock());
    lastGreenLight = clock();
  } else {
    return; //hasn't been 1sec since last GreenLight
  }
  
  char buf[headerSize];
  memset(buf, 0, sizeof(buf));
  
  //header
  buf[iMsgType] = mtReady;
  buf[iGameID] = giIn->gameID;
  buf[iPlayerNo] = giIn->playerNo;

  string home("home");
  if(bossIP == home){
    return; //I don't have to signal myself... do I?
  }

  for( map<string, P2pNodeInfo *>::iterator lIterAddr = myNI.begin();
      lIterAddr != myNI.end();
      lIterAddr++){
    string addr = lIterAddr->first;
    if(addr == bossIP){
      basicSend(buf, headerSize, addr);
      break;
    }
  }
	
	
  
  
}

void MessageManager::signalGo(){
	if(dataSize <0)
		dataSize = thisGI->inputSize;
	
  char buf[headerSize + 2];
  memset(buf, 0, sizeof(buf));
  unsigned int localMaxDelay = getLocalMaxDelay();
	DPRINT(0,"Signaling go to everybody...\n");
  //header
  buf[iMsgType] = mtGo;
	
  
	outSockSem->enterCritical();
	

  for(map<string, P2pNodeInfo *>::iterator lIterAddr = myNI.begin();
      lIterAddr != myNI.end();
      lIterAddr++){

    string addr = lIterAddr->first;
		unsigned int hisWaitTime = localMaxDelay - myNI[ addr ]->estDelay;
		
		packInBytes(hisWaitTime, B_HOLD_TIME, buf + iData);
		DPRINT(0,"mr %s should wait for %ums\n", addr.c_str(), hisWaitTime);
		basicSend(buf, headerSize + 2, addr, false);
  }
  outSockSem->exitCritical();
	DPRINT(0,"mr MYSELF should wait for %ums\n",localMaxDelay);
	if(calleeInstance)
		calleeInstance->hold( localMaxDelay );
	
}

void MessageManager::receivedPing(char *inBuf, string addrIn){
	unsigned int hisLastPingTime =0;
	char outType;
	switch( inBuf[iPingType]){
		case TIME_PING:
			{
			int hisMaxDelay = atoi(inBuf+iData);
			hisLastPingTime = atoi(inBuf + iData + INT_DIG);
			DPRINT(0,"%s says his maxDelay is %d\n", addrIn.c_str(), hisMaxDelay );
			newDelay(addrIn, -1, hisMaxDelay); //haven't calculated delay
			}
			outType = TIME_PING;
			break;
		case STRESS_PING:
			outType = STRESS_PING;
			//just pong it
			break;
		default:
			DPRINT(0,"This is not happening\n");
			outType = (char)-1;
			break;
	}
 

  char szBuffer[ headerSize + 2*INT_DIG];
  memset( szBuffer, 0, sizeof(szBuffer) );

  szBuffer[iMsgType] = mtPong;
	szBuffer[iPlayerNo] = thisGI->playerNo; //let's be kind
	szBuffer[iPingType] = outType;
  sprintf(szBuffer+iData, "%d", myMMInfo->maxDelay);
  sprintf(szBuffer+iData+INT_DIG, "%u", hisLastPingTime);
   

  //DPRINT(0,"Ponging %u\n", hisLastPingTime);
  basicSend(szBuffer, headerSize + 2*INT_DIG, addrIn);
	
	
}


void MessageManager::receivedPong(char *inBuf, string addrIn){
	if( inBuf[iPingType] == STRESS_PING ){
		myNI[addrIn]->receivedPongs++;
		return;
	}
  int hisMaxDelay = atoi(inBuf+iData); 
  clock_t hisLastPingTime = atoi(inBuf + iData + INT_DIG);
  if(hisLastPingTime != myNI[addrIn]->lastPingTime){
    DPRINT(0,"Discarding ping %u, expected %u\n",
        hisLastPingTime, myNI[addrIn]->lastPingTime );
		myNI[addrIn]->greetSpace *=3;
		myNI[addrIn]->greetSpace /=2;
    return;
  }
  myNI[addrIn]->lastPongTime = clock();
  
  int estDelay = (myNI[addrIn]->lastPongTime
      - myNI[addrIn]->lastPingTime)
      *1000/CLOCKS_PER_SEC
			/2; //roundtrip
      
  newDelay( addrIn, estDelay, hisMaxDelay );
	if( myNI[addrIn]->playerNo < 0 )
		myNI[addrIn]->playerNo = inBuf[iPlayerNo];
}

void MessageManager::receivedData(char *inBuf, string addrIn){
  int frameNo = getPackedInBytes(inBuf + iFrameNo, B_FRAME_NO);
  int playerNo = inBuf[iPlayerNo];
  int recvDataSize = inBuf[iDataSize];
  
  
  unsigned char *myBuf = new unsigned char[dataSize];
  for(int i=0; i < recvDataSize; i++)
    myBuf[i] = inBuf[iData+i]; //bits are the same, unless dealing with different endian machines
  //store it and wait for confirmation
  
	
	struct gameInfo *gi = (struct gameInfo *)malloc( sizeof(struct gameInfo) );
	gi->frameNo = frameNo;
	gi->playerNo = playerNo;
	
	int syncDiff = calleeInstance->receivedData( myBuf, dataSize, gi );
	
	int msNeeded = myNI[ addrIn ]->estDelay*2 //ack there, ok here... data already here
									- syncDiff*MS_PER_SYNC;
	DPRINT(0,"I will need %dms to input this correctly (2x(%dms) - (%d frames))\n",
		msNeeded,
		myNI[ addrIn ]->estDelay,
		syncDiff);
	//spare ms == -(needed ms)
	
	//let's ack him
	char outBuf[headerSize + B_FRAME_NO];//spare bytes in case of ack
	for(int i=0;i<headerSize + B_FRAME_NO;i++)
		outBuf[i] = 0;
	int outSize;
	int key; 
	if( syncDiff > 0 ){ //can't be on "last processed frame" (==)
		DPRINT(0,"Correct input:%d, acking\n", frameNo);
		
		outSize = headerSize;
		key = makeSyncAndPlayerNoKey( frameNo, playerNo);
		outBuf[iMsgType] = mtAckData;
	}
	else {
		//p2ps::rd packed it.. thanx
		int bounceFrameNo = gi->frameNo;
		DPRINT(0,"Bouncing %d to %d\n", frameNo, gi->frameNo);
		
		outSize = headerSize + B_FRAME_NO;
		key = makeSyncAndPlayerNoKey( bounceFrameNo, playerNo);
		outBuf[iMsgType] = mtNoAckData;
		
		//...and pack bounceFrame
		packInBytes( bounceFrameNo, B_FRAME_NO, outBuf+iData);
		
		//queueInTime(mensaje, gi->frameNo + syncDiff);
	}
	DPRINT(0,"Received data @synco:%u, creating msg (size:%d)on %d\n", thisGI->frameNo, dataSize, key);
	//correct or bounce frame... create it either way
	msgInProcess[key] = new P2pMessage( myBuf, dataSize, thisGI->totalPlayers, frameNo);
	
	packInBytes(frameNo, B_FRAME_NO, outBuf + iFrameNo);
	outBuf[iPlayerNo] = thisGI->playerNo;
	//do NOT use other than outSize
	basicSend( outBuf, outSize, addrIn);
	

  delete[] myBuf;
	free( gi );
}

void MessageManager::receivedClosedConnection(string addrIn){
  remove(addrIn);
}

void MessageManager::receivedPlayerReady(char *inBuf){
  struct gameInfo *gameInfo;
  gameInfo = (struct gameInfo *)malloc(sizeof(struct gameInfo));
  gameInfo->playerNo = inBuf[iPlayerNo];
  gameInfo->gameID = inBuf[iGameID];
  
  if(calleeInstance)
    calleeInstance->receivedPlayerReady(gameInfo);
	else
		DPRINT(0,"No calleeInstance!!");
  free(gameInfo);
}


void MessageManager::receivedGo(char *inBuf){
	if( thisGI->greenLighted ){
		DPRINT(0,"Discarding greenLight\n");
		return;
	}
		
	int holdTime = 0;
	holdTime = getPackedInBytes(inBuf + iData, B_FRAME_NO);
	
	
  if(calleeInstance){
		calleeInstance->hold( holdTime );
    calleeInstance->receivedGo();
	}
}


void MessageManager::receivedAckData( char *inBuf, string addrIn ){
	int frameNo = getPackedInBytes(inBuf + iFrameNo, B_FRAME_NO);
	int playerNo = inBuf[iPlayerNo];
	int key = makeSyncAndPlayerNoKey( frameNo, thisGI->playerNo);
	
	
	unsigned int deltaT = clk2ms(clock() - myNI[addrIn]->lastSentTime);
	DPRINT(0,"Ack from %dp%d after %ums of waiting\n", frameNo, playerNo, deltaT);
	if( msgInProcess.find(key) == msgInProcess.end() ){
		DPRINT(0,"No such msg... did you deleted it? (maybe from a noAck)\n");
		char *myBuf = new char[headerSize];
		
		myBuf[iMsgType] = mtKillData;
		myBuf[iPlayerNo] = thisGI->playerNo;
		packInBytes(frameNo, B_FRAME_NO, myBuf + iFrameNo );
		basicSend(myBuf, headerSize, addrIn);
		
		
		
		return;
	}
	
	msgInProcess[key]->playerConfirm( playerNo );
	
	if(msgInProcess[key]->isReady())
		messageCollected(key);
}
void MessageManager::messageCollected(int key){
	DPRINT(0,"Key %d ready\n",key);
	int bounceFrame = msgInProcess[key]->getFrameBounce();
	
	char myBuf[headerSize+B_FRAME_NO];//TODO: resize this
	for(int i=0; i<headerSize; i++)
			myBuf[i] = 0;
	myBuf[iPlayerNo]	=  thisGI->playerNo;
	
	struct gameInfo *gI = (struct gameInfo *)malloc( sizeof(struct gameInfo) );
	int frameNo = getFrameNo(key);
	if( bounceFrame < 0 ){
		DPRINT(0, "MM: \"All\" confirmed for %d...", frameNo);
		++myMMInfo->sentMsgCount;
		packInBytes(myMMInfo->sentMsgCount, B_FRAME_NO,  myBuf + iMsgCount);
		myBuf[iMsgType] = mtOkData;
		
		packInBytes(frameNo, B_FRAME_NO, myBuf + iFrameNo);
		
		gI->frameNo = frameNo;
		gI->playerNo = thisGI->playerNo; //someone confirmed MY message
		
		if( calleeInstance )
			calleeInstance->confirmMsg( gI );
		
		outSockSem->enterCritical();
		for(  map<string, P2pNodeInfo *>::iterator lIter = myNI.begin();
        lIter != myNI.end();
        lIter++){

			string out = lIter->first;
			basicSend(myBuf, headerSize, out, false);
		}
		outSockSem->exitCritical();
		
		//TODO:if it couldn't make it, don't count it
	}
	else{
		
	
		myBuf[iMsgType]		= mtBounce;
	
		packInBytes( frameNo, B_FRAME_NO, myBuf+iFrameNo);
		packInBytes( bounceFrame, B_FRAME_NO, myBuf+iData);
		//same frameNo
		
		gI->frameNo = frameNo;
		gI->playerNo = thisGI->playerNo; //someone UNconfirmed MY message
	
		int time2wait = (bounceFrame - thisGI->frameNo)*MS_PER_SYNC;
		//tell people
		for(  map< string, P2pNodeInfo *>::iterator lIterAddr = myNI.begin();
          lIterAddr != myNI.end();
          lIterAddr++){
			string out = lIterAddr->first;
			int delta = time2wait - myNI[out]->estDelay;
			if(delta < 0)
				delta = 0;
			sendLater(myBuf, headerSize+B_FRAME_NO, out, delta);
		}
		//tell here
		calleeInstance->confirmBounce(thisGI->playerNo, frameNo, bounceFrame);
		//calleeInstance->deConfirmMsg( gI );
	
	}
	free( gI );
	
	
	//TODO:this below should be in critical sec?
	
	delete msgInProcess[key];
	msgInProcess.erase( msgInProcess.find(key) );
		
}
void MessageManager::receivedOkData(char *inBuf, string addrIn){
	
	int frameNo = getPackedInBytes(inBuf + iFrameNo, B_FRAME_NO);
	int playerNo = inBuf[iPlayerNo];
	int key = makeSyncAndPlayerNoKey( frameNo, playerNo);
	
	int hisMsgCount = getPackedInBytes(inBuf+iMsgCount, B_FRAME_NO);
	if( msgInProcess.find(key) == msgInProcess.end() ){
		DPRINT(0,"receivedOkData:No such msg... you should worry \n");
		return;
	}
	
	myNI[addrIn]->recvMsgCount++;
	
	DPRINT(0,"Ok on %dp%d, msg count for mr %s: %u(mine), %u(his)\n",
			frameNo,
			playerNo,
			addrIn.c_str(),
			myNI[addrIn]->recvMsgCount,
			hisMsgCount);
	
	
	struct gameInfo *gameInfo = (struct gameInfo *)malloc( sizeof(struct gameInfo) );
	gameInfo->frameNo = frameNo;
	gameInfo->playerNo = playerNo;
	
	
  
	
	if( calleeInstance )
		calleeInstance->confirmMsg( gameInfo );
	
	//this input is no longer useful
	delete msgInProcess[key];
	msgInProcess.erase( msgInProcess.find(key) );
	
	free(gameInfo);
}
void MessageManager::receivedBounce( char *inBuf ){
	
	int originalFrame = getPackedInBytes( inBuf+iFrameNo,B_FRAME_NO);
	int bounceFrame = getPackedInBytes(inBuf+iData, B_FRAME_NO);
	int player = inBuf[iPlayerNo];
	
	calleeInstance->confirmBounce(player, originalFrame, bounceFrame);
	
}

void MessageManager::receivedLoadState( char *inBuf ){
	int frameSaved, frameLoading;
	int hisPlayerNo, loadedPlayerNo;
	
	//header
	//inBuf[iPlayerNo];
	frameSaved = getPackedInBytes(inBuf + iFrameNo, B_FRAME_NO);
	//data
	loadedPlayerNo = inBuf[iData];
	frameLoading = getPackedInBytes(inBuf + iData +1, B_FRAME_NO);
	
	
	calleeInstance->queueStateLoading(frameSaved, frameLoading, loadedPlayerNo);
	//we're losing the player with delay!
		
}

void MessageManager::receivedNoAckData( char *inBuf, string addrIn){
	int originalFrame = getPackedInBytes(inBuf + iFrameNo, B_FRAME_NO);
	int playerNo = inBuf[iPlayerNo];
	int key = makeSyncAndPlayerNoKey( originalFrame, thisGI->playerNo);
	
	unsigned int deltaT = clk2ms(clock() - myNI[addrIn]->lastSentTime);
  DPRINT(0,"noAck from %dp%d after %ums of waiting\n", originalFrame, playerNo, deltaT);
	if( msgInProcess.find(key) == msgInProcess.end() ){
		DPRINT(0,"No such msg... did you deleted it? (maybe from a noAck)\n");
		return;
	}
	
	int bounceFrame = getPackedInBytes(inBuf+iData, B_FRAME_NO);
	
	msgInProcess[key]->playerDeConfirm(playerNo, bounceFrame);
	
	if( msgInProcess[key]->isReady() )
		messageCollected(key);
	
} 
void MessageManager::receivedKillData( char *inBuf ){
	int frameNo = getPackedInBytes(inBuf + iFrameNo, B_FRAME_NO);
	int playerNo = inBuf[iPlayerNo];
	int key = makeSyncAndPlayerNoKey( frameNo, playerNo);
	
	//unsigned int hisMsgCount = atoi(inBuf+iMsgCount);
	DPRINT(0,"Kill on %up%d\n",	frameNo,	playerNo);
	
	struct gameInfo *gameInfo = (struct gameInfo *)malloc( sizeof(struct gameInfo) );
	gameInfo->frameNo = frameNo;
	gameInfo->playerNo = playerNo;
	
	
	calleeInstance->deConfirmMsg( gameInfo );
	if( msgInProcess.find(key) != msgInProcess.end() ){
		delete msgInProcess[key];
		msgInProcess.erase( msgInProcess.find(key) );
	} else
		DPRINT(0,"Couldn't kill %d... should I be worried?\n",key);
	
	free(gameInfo);
	
	
}
void MessageManager::receivedStress(){
	/*
	stressConns();
	*/
}



int MessageManager::getLocalMaxDelay(){
	int out = 0;
	for( map<string, P2pNodeInfo *>::iterator iter = myNI.begin();
          iter != myNI.end();
          iter++){
      if( ((*iter).second)->estDelay >= 0)
        out = (out >((*iter).second)->estDelay)?out:
          ((*iter).second)->estDelay;


    }
	return out;
}

void MessageManager::newDelay(string addrIn, int estDelay,int hisMaxDelay){
  
  if(estDelay == 0 )
    estDelay = 1;
  if(hisMaxDelay == 0)
    hisMaxDelay = 1;//let's round it up

  P2pNodeInfo *hisNI = myNI[ addrIn ];
  int byWeight,maxDelay=-1;
   
  outSockSem->enterCritical(); //two pings could arrive very close to each other
	
  if(estDelay > 0){ 
    //change one's delay
    if( hisNI->estDelay < 0)
      hisNI->estDelay = estDelay;
		if( hisNI->estDelay > estDelay)
			byWeight = myMMInfo->loWeight;
		else
			byWeight = myMMInfo->hiWeight;
		
    hisNI->estDelay *= (byWeight-1);
    hisNI->estDelay += estDelay;
    hisNI->estDelay /= byWeight;
    DPRINT(0,"Changin mr %s's delay to %u\n", addrIn.c_str(), hisNI->estDelay);

    //recalculate our global delay
		maxDelay = getLocalMaxDelay();
  }
  
  maxDelay = (maxDelay > hisMaxDelay)?maxDelay:hisMaxDelay;
  myMMInfo->maxDelay = maxDelay;

  if(calleeInstance)
    calleeInstance->setNewDelay(maxDelay*3); //data, ack, ok

 

  

  outSockSem->exitCritical();     
}
void MessageManager::getMaxDelay(){
  clock_t chkpt = clock();
  DPRINT(0,"Request for new delay calculation @ %u\n",chkpt);

  for(int i=0; i<N_ESTIMATIONS; i++){
		chkpt = clock();
    bool allClear = false;
    while(!allClear){
			greet();
			if( !myMMInfo->looping)
				tryAndReceive(false);
			allClear = true;
			for( map<string, P2pNodeInfo *>::iterator iter = myNI.begin();
          iter != myNI.end();
          iter++){
        if( ((*iter).second)->lastPongTime < chkpt){ 
          allClear = false;
          break;
  			}
			}
    }
  }
  DPRINT(0,"delay calculated\n");
}
void MessageManager::receivedPkg(char *inBuf, string addrIn){
	DPRINT(0,"mr %s ", addrIn.c_str());
	if( myMMInfo->allConnected && myNI.find(addrIn) == myNI.end() ){
		DPRINT(0,"invalid recvd pkg\n");
		return; //invalid addr
	}
	switch( inBuf[iMsgType] ){
		case mtOpenConn:
			DPRINT(0,"is opening a connection\n");
			receivedOpenConnection(addrIn);
			break;

		case mtHandShake:
			DPRINT(0,"is confirming a connection\n");
			receivedHandShake(addrIn);
			break;

		case mtPing:
			//DPRINT(0,"just pinged me\n");
			receivedPing(inBuf, addrIn);
			break;

		case mtPong:
			//DPRINT(0,"just ponged me\n");
			receivedPong(inBuf, addrIn);
			break;

		case mtData:
			DPRINT(0,"sent me data\n");
			receivedData(inBuf, addrIn);
			break;
		
		case mtLoadState:
			DPRINT(0,"wants to load state\n");
			receivedLoadState( inBuf );
			break;

		case mtCloseConn:
			DPRINT(0,"is closing his connection\n");
			receivedClosedConnection(addrIn);
			break;
			
		case mtReady:
			DPRINT(0,"is ready\n");
			receivedPlayerReady(inBuf);
			break;
			
		case mtGo:
			DPRINT(0,"is saying we should start emu\n");
			receivedGo(inBuf);
			break;
			
		case mtAckData:
			DPRINT(0,"Just acked me\n");
			receivedAckData(inBuf, addrIn);
			break;
			
		case mtOkData:
			DPRINT(0,"Just OK'd me\n");
			receivedOkData(inBuf, addrIn);
			break;
			
		case mtNoAckData:
			DPRINT(0,"Just noAcK'd me\n");
			receivedNoAckData( inBuf, addrIn);
			break;
			
		case mtKillData:
			DPRINT(0,"Just KILL'd me\n");
			receivedKillData( inBuf );
			break;
			
		case mtBounce:
			DPRINT(0,"just confirmed bounce");
			receivedBounce(inBuf);
			break;
			
		case mtStress:
			DPRINT(0,"just told me to stress \n");
			receivedStress();
			break;
			
		default:
			DPRINT(0,"sent me an empty package\n");
			//error
			break;

	}
	//DPRINT(0,"...finished processing package\n");

}
void MessageManager::stressConns(){
	int nNodes = myNI.size();
	int totalBombing = N_STRESS_LOAD * nNodes;
	DPRINT(0,"Stressing...\n");
	//pkg
	char buf[headerSize + 2*INT_DIG];
	memset( buf, 0, sizeof(buf) );
	buf[iMsgType] = mtPing;
	buf[iPingType] = STRESS_PING;
	map<string, P2pNodeInfo *>::iterator iAddr
					= myNI.begin();
	
	for(int i=0; i<totalBombing; i++, iAddr++){
		if( iAddr == myNI.end() )
			iAddr = myNI.begin();
		
		basicSend( buf, headerSize, iAddr->first );
		if( !myMMInfo->looping)
				tryAndReceive(false);
		
	}
	
	DPRINT(0,"Number of received pongs:\n");
	for( iAddr = myNI.begin(); iAddr != myNI.end(); iAddr++)
		DPRINT(0,"%s:%d\n", iAddr->first.c_str(), iAddr->second->receivedPongs);
	
	Sleep( getLocalMaxDelay() );
}


void MessageManager::stressAll(){
	//called by boss
	/*
	stressConns();
	basicSend(player)
	create msg
	waitfortest
	continue
	*/
	DPRINT(0,"WARNING:No stress testing since haven't implemented passing msg\n");
}

void MessageManager::retrial(){
	
}

