#ifndef MSGMANAGER
#define MSGMANAGER
#include "CalleeBack.h"
#include "p2pSync.h"
#include "P2pSem.h"
#include "p2pNodeInfo.h"
#include <time.h>

/*STL includes*/
#include <list>
#include <queue>
#include <map>
#include <string>

/*defs*/
#define DEFAULT_GAMEID 215
#define CONNECT_TO_S 3
#define SELECT_TO_US 500 //0.5ms
#define IP_CSTR_SIZE 16 //123.567.901.345 and '\0'
#define INT_DIG 8 //7 digits + '\0'

/* bytes used/necessary */
#define B_HOLD_TIME 2
#define B_FRAME_NO 3

#ifndef WListeningSocket_VERSION
  #define WListeningSocket_VERSION MAKEWORD(2,2)
#endif
/* initialized w/MM */
struct MMInfo {
  int maxDelay;
  int loWeight; //weight for new delays lower than stored
	int hiWeight; //weight for new delays higher than stored
  bool looping;
	unsigned int sentMsgCount; //ok'd data msgs
	bool allConnected;
};

struct msgInfo {
  unsigned int sentTime; //to compare with time when ack was received

};


class CalleeBack;
struct gameInfo;
class MessageManager
{


  public:
		
		virtual ~MessageManager();
    void greet();
    void broadcastData(unsigned char *msg, int msgSize, int frameNo);
		void broadcastStateLoad( int syncLoaded, int syncLoading, int playerNo);
    void signalReady(struct gameInfo *giIn);
    void signalGo();
    void setCallbacks(CalleeBack *callee);
    void getMaxDelay();
		void stressAll();
  protected:
		void receiveLoop();
		void retrialLoop();
		MessageManager(CalleeBack *calleeback);
    std::string bossIP; //he's the one who gives green light and other stuff
    bool invite(std::string ip4addr);
    bool isStored(std::string strin);
		
		void remove(std::string strin);
    bool looping;
    void messageCollected(int key);
  protected:
		struct gameInfo *thisGI;//sign messages w/playerNo, gameID & totalPlayers
    struct MMInfo *myMMInfo; //common info
    std::map <std::string, P2pNodeInfo *> myNI; //xxx.xxx.xxx.xxx:XXXX -> time
		CalleeBack *calleeInstance;
    P2pSem *outSockSem;
    std::map<int, P2pMessage *> msgInProcess;
		std::map<std::string, int> ackedFromStress;
		void connectUp(std::queue<std::string> IPsToConnectTo, unsigned int targetConns);
		void receivedPkg(char *inBuf, std::string addr);
    void receivedOpenConnection(std::string addr);
    void receivedHandShake(std::string addr);
    void receivedPing(char *recvBuf, std::string addr);
    void receivedPong(char *inBuf, std::string addr);
    void receivedData(char *inBuf, std::string addr);
    void receivedClosedConnection(std::string addr);
    void receivedPlayerReady(char *inBuf);
    void receivedGo(char *inBuf);
    void receivedAckData(char *inBuf, std::string addr);
    void receivedOkData(char *inBuf, std::string addr);
		void receivedLoadState( char *inBuf );
		void receivedNoAckData( char *inBuf, std::string addr);
		void receivedKillData( char *inBuf );
		void receivedBounce(char *inBuf);
		void receivedStress();
    bool connect();
    void newDelay(std::string addrIn, int calculatedDelay,int hisMaxDelay);
		int getLocalMaxDelay();
		void stressConns();
		void retrial();
  protected:
    enum pkgIndexes{
			iMsgCount,
			iMsgType = iMsgCount + B_FRAME_NO,
			iGameID,
			iPingType,
			iFrameNo = iPingType,//needs 3 bytes, reusing pingType
			iPlayerNo=iFrameNo+B_FRAME_NO,
			iDataSize,
			//iData & headerSize should be the last "unnamed"
			headerSize, 
			iData = headerSize
		};
		int dataSize;
    enum msgTypes{
			mtOpenConn=112,//70
			mtHandShake,
			mtPing,//72
			mtPong,
			mtData,
			mtCloseConn,//75
			mtReady,//ready to start emu'd machine
			mtGo,//start emu'ing machine
			mtAckData,
			mtNoAckData, //didn't arrive ok
			mtOkData, //data you already received should go
			mtKillData, //data you already received, but shouldn't go
			mtLoadState,
			mtBounce,
			mtStress
		};
	protected:	
		virtual void basicSend( char *buf, int size, std::string addr, bool block = true)=0;
    virtual int tryAndReceive(bool shortWait)=0;
		virtual void sendLater(char *buf, int size, std::string addr, int deltaInMs)=0;
		virtual void store(std::string strin)=0; //! probably you'll need some more info on nodes


};

#endif

