#include "P2pPlayer.h"
#include <assert.h>
#include <windows.h>
#ifdef __MINGW32__
  //#define DPRINT(format, ...) if(1)bprintf(0, format, ## __VA_ARGS__ )
	#define DPRINT if(1)bprintf
  extern int (__cdecl *bprintf) (int nStatus, TCHAR* szFormat, ...);
#else
  #define DPRINT(format, ...)
#endif

//all concurrent access are managed outside
using namespace std;

P2pPlayer::P2pPlayer(int playerNoIn, struct gameInfo *GIin){
  playerNo = playerNoIn;
  inputCache = NULL;
  setInputCache = NULL;
  ready = false;
  myGI = GIin;
}

P2pPlayer::~P2pPlayer( ){
  if(!inputCache)
    free(inputCache);
  //iterate over map to clean it
}


bool P2pPlayer::setInput( unsigned char *inBuf, int size, int frameNo){
  if( inputCache == NULL || setInputCache == NULL ){
		inputCache = new unsigned char[size];
    setInputCache = new unsigned char[size];
		
		for(int i=0; i< size; i++){
			inputCache[i]    = 0;
			setInputCache[i] = 0;
		}
		
		
  }
	
	bool diff = false;
	for(int i = 0; i < myGI->inputSize; i++)
		if( inBuf[i] != setInputCache[i] ){
			diff = true;
			break;
		}
	
	//assert( inputCache != NULL );
	
	if( !diff ) //if it isn't different, don't change anything
		return false;
	
	//create new p2pMsg from candidate
	P2pMessage *msgIn = new P2pMessage( inBuf, size, myGI->totalPlayers, frameNo);
	
  if( inputMsgCache.find(frameNo) != inputMsgCache.end()
	){
		//let's take it out for comparison
		map< int, P2pMessage *>::iterator iter =
      inputMsgCache.find(frameNo);
    
		
		//let's clear storage
		delete inputMsgCache[frameNo];
    inputMsgCache.erase( iter );
		
		DPRINT(0,"pp:deleted previous input on %d\n", frameNo);
  }
  //let us hope that the number of players have been set
  //assert( myGI->totalPlayers > 0);
  //assert( myGI->totalPlayers <= 8);

	DPRINT(0,"Creating message on %d\n", frameNo);
  //copy here
  inputMsgCache[frameNo] = msgIn;
	
  

  //cache it
  for(int i=0; i < size; i++)
    setInputCache[i] = inBuf[i];
  

  //if we had to cache it, then it was different
  return diff;//true
  
}

int P2pPlayer::getInput( unsigned char *outBuf, int frameNo){
	if( inputCache == NULL || setInputCache == NULL ){
		inputCache = new unsigned char[myGI->inputSize];
    setInputCache = new unsigned char[myGI->inputSize];
		
		for(int i=0; i< myGI->inputSize; i++){
			inputCache[i]    = 0;
			setInputCache[i] = 0;
		}
	}
  //memset(outBuf, 0, inputSize);

	if( inputMsgCache.size() > 0 ){
		DPRINT(0, "Player: frameNo:%d : Have %d on queue\n", frameNo, inputMsgCache.size() );
	}
	if( bounceInputMsgCache.size() > 0){
		DPRINT(0, "Player: frameNo:%d : Have %d on BOUNCE queue\n", frameNo, bounceInputMsgCache.size() );
	}
	
  if( inputMsgCache.find( frameNo ) != inputMsgCache.end()
			//&& inputMsgCache[frameNo]->isReady() ){
			// above condition is checked separately
		){
    //oh, let's update the output cache
		DPRINT(0,"New input on %u\n", frameNo);
    inputMsgCache[frameNo]->getInput( inputCache );
		lastOriginalFrameProcessed = frameNo;
  }
	if( bounceInputMsgCache.find( frameNo ) != bounceInputMsgCache.end()	){
		// get range of elements under frameNo
		pair< multimap<int, P2pMessage *>::iterator, 
					multimap<int, P2pMessage *>::iterator> range;
		range = bounceInputMsgCache.equal_range(frameNo);
		// get highest originalFrame and  reference
		P2pMessage *selected = range.first->second;
		int highestOriginalFrame=selected->getOriginalFrame();
		
		for( multimap<int, P2pMessage *>::iterator i = range.first;
					i != range.second;
					i++){
			if( i->second->getOriginalFrame() > highestOriginalFrame ){
				// select it
				selected = i->second;
				highestOriginalFrame = selected->getOriginalFrame();
			}
		}
		
		//input hasn't changed since original frame
		if( selected->getOriginalFrame() > lastOriginalFrameProcessed ){
		
			DPRINT(0,"New bounce input on %u (original is %d)\n",
						frameNo, selected->getOriginalFrame() );
			selected->getInput( inputCache );
			lastOriginalFrameProcessed =
					selected->getOriginalFrame();
		}
	}
	for(int i=0; i< myGI->inputSize; i++)
		outBuf[i] = inputCache[i];
	
	return myGI->inputSize;
}

int P2pPlayer::getSetCache( unsigned char *outBuf ){
  if(myGI->inputSize <= 0)
    return -1;
  for(int i=0; i < myGI->inputSize; i++)
    outBuf[i] = inputCache[i];
		
	return myGI->inputSize;
}


void P2pPlayer::clearInput(int frameNo){
  //find on map
	map<int, P2pMessage *>::iterator result = inputMsgCache.find( frameNo );
  if( result != inputMsgCache.end() ){
    
		DPRINT(0,"Trying to delete frameNo:%u\n",frameNo);
		//delete pp
		delete inputMsgCache[frameNo];
		//erase()
		inputMsgCache.erase( result );
	}
	
	
	//bounce section
	
	result = bounceInputMsgCache.find(frameNo);
	if( result != bounceInputMsgCache.end() ){
    
		DPRINT(0,"Trying to delete %d BOUNCEframeNos:%u\n",bounceInputMsgCache.count(frameNo),frameNo);
		//delete pp
		pair< multimap<int, P2pMessage *>::iterator, 
					multimap<int, P2pMessage *>::iterator> range;
		range = bounceInputMsgCache.equal_range(frameNo);
		for( multimap<int, P2pMessage *>::iterator i = range.first;
					i != range.second;
					i++){
			delete i->second;
		}
		//erase()
		bounceInputMsgCache.erase( range.first, range.second );
	}
}

void P2pPlayer::confirmMsg(int frameNo){
	if( inputMsgCache.find(frameNo) != inputMsgCache.end() ){
		inputMsgCache[frameNo]->fullConfirm();
		return;
	} else {
		DPRINT(0,"player:Whoa! %u\n", frameNo);
	}
	
	
}
void P2pPlayer::confirmBounce( int originalFrame, int bFrameNo ){
	bool bounced = false;
	
	
	if( inputMsgCache.find( originalFrame ) != inputMsgCache.end() ){
		DPRINT(0,"you should confirming, not bouncing... sigh\n");
		bounceInput(originalFrame, bFrameNo);
		confirmBounce(originalFrame, bFrameNo);
		bounced = true;
	}
	multimap<int, P2pMessage *>::iterator iter;
	
	if(!bounced){
	
		for(iter = bounceInputMsgCache.begin(); iter != bounceInputMsgCache.end(); iter++){
			
			if( (iter->second)->getOriginalFrame() == originalFrame ){
				P2pMessage *curr = iter->second;
				bounceInputMsgCache.erase(iter);//remove from previous position
				bounceInputMsgCache.insert(
						make_pair( bFrameNo, curr) );//move to new bFrame
				curr->fullConfirm();
				bounced = true;
				break;
			}
			
		}
	
	}
	
	if(!bounced)
		DPRINT(0,"pplay: ERROR! couldn't bounce");
	else
		DPRINT(0,"Finished bouncing\n");
}
void P2pPlayer::deConfirmMsg(int frameNo){
	map<int, P2pMessage *>::iterator result = inputMsgCache.find(frameNo);
	if( result == inputMsgCache.end() ){
		DPRINT(0,"player:dude! %u\n", frameNo);
		return;
	}
	delete inputMsgCache[frameNo];
	inputMsgCache.erase( result );
}

int P2pPlayer::inputReady( int frameNo ){
	//first hand input
	if( inputMsgCache.find( frameNo ) != inputMsgCache.end() ){
		if( !inputMsgCache[frameNo]->isReady() )
			return frameNo; //new input NOT ready
		else
			return -1; //new input ready
	}
	//bounce input
	if( bounceInputMsgCache.find( frameNo ) != bounceInputMsgCache.end() ){
	  pair< multimap<int, P2pMessage *>::iterator, 
					multimap<int, P2pMessage *>::iterator> range;
		range = bounceInputMsgCache.equal_range(frameNo);
		int highestOriginalFrame=-1; //-1 == everything ok
	  for( multimap<int, P2pMessage *>::iterator i = range.first;
					i != range.second;
					i++){
			if( 	!i->second->isReady()
						&& i->second->getOriginalFrame() > highestOriginalFrame )
				highestOriginalFrame = i->second->getOriginalFrame();
			
		}
		
		return highestOriginalFrame;
		
	}
	
	return -1; //no new input of any kind
}

void P2pPlayer::bounceInput( int oldInput, int newInput){
	map<int, P2pMessage *>::iterator result = inputMsgCache.find(oldInput);
	if(result != inputMsgCache.end()){
		P2pMessage *bounced = inputMsgCache[oldInput];
		inputMsgCache.erase(result);
		bounceInputMsgCache.insert( make_pair(newInput,bounced) );
	}else{
		DPRINT(0,"pp: Couldn't bounce orig:%d, no such frame", oldInput);
	}
}

