#include "WinMessageManager.h"
#include "WinMessage.h"
#ifdef _MSC_VER
  #pragma comment(lib, "winmm.lib")
  #pragma comment(lib, "wsock32.lib")
  #pragma comment(lib, "advapi32.lib")
  #pragma comment(lib, "ws2_32.lib")
#endif

#ifdef __MINGW32__
  //#define DPRINT(format, ...) if(1)bprintf(0, format, __VA_ARGS__ )
	//#define DPRINT(format, args...) if(1)bprintf(0, ## args)
	#define DPRINT if(1)bprintf
  extern int (__cdecl *bprintf) (int nStatus, TCHAR* szFormat, ...);
#else
  #define DPRINT(format, ...)
#endif

#define MAIN_PORT 27815
#define REDUNDANT_STEP 2

DWORD WinMessageManager::receiveThreadStart(void *param){
	((WinMessageManager *)param)->receiveLoop();
	return 0;
}


DWORD WinMessageManager::redundantThreadStart(void *param){
	DPRINT(0,"redundant loop started\n");
	((WinMessageManager *)param)->redundantLoop();
	return 0;
}


bool WinMessageManager::initialized = false;

using namespace std;

WinMessageManager::WinMessageManager(list<string> lIPaddr, CalleeBack *callback)
: MessageManager(callback){
	
	list<string>::iterator lIter = lIPaddr.begin();
  int goalSize = lIPaddr.size() -1;//we want to have all connected
  size_t myIndex = 0; //my position in the list
  queue<string> qConnectTo;

	mainSock = INVALID_SOCKET;
  while( lIter != lIPaddr.end() ){
    list<string>::iterator current=lIter;
    lIter++;


    myIndex++;

    if(*current == "home")
      break;

  }
  


  DPRINT(0,"I should invite ");
  while(lIter != lIPaddr.end() ){
    string current=*lIter;
    lIter++;

    qConnectTo.push(current);
    DPRINT(0,"%s, ",current.c_str());
  }
  DPRINT(0,"... I should create %d conns\n",goalSize);
	
	if( initialized || !initWSocks() || !initListening() ){
    DPRINT(0,"Listening Socket not initialized\n");
  } else
		initialized = true;
	
  connectUp(qConnectTo, goalSize);
  bossIP = *(lIPaddr.begin());
  DPRINT(0,"Boss is \"%s\"\n",(*(lIPaddr.begin())).c_str());
  DWORD receiveThreadID;
	HANDLE otherThread=CreateThread(
                     NULL,       // default security attributes
                     0,          // default stack size
                     (LPTHREAD_START_ROUTINE) receiveThreadStart,
                     this,
                     0,          // default creation flags
                     &receiveThreadID);
  
  if(otherThread == NULL){
		//throw
    DPRINT(0,"Couldn't create receive thread\n");
  }
	
	
	redundantLooping = false;
	reduntSem = new P2pSem();
	
	otherThread=CreateThread(
                     NULL,       // default security attributes
                     0,          // default stack size
                     (LPTHREAD_START_ROUTINE) redundantThreadStart,
                     this,
                     0,          // default creation flags
                     &receiveThreadID);
	
	if(otherThread == NULL){
		//throw
    DPRINT(0,"Couldn't create redundant thread\n");
  }
}


bool WinMessageManager::initWSocks(){
	DPRINT(0,"Initializing wsocks\n");
  WORD	wWinsockVersionRequested;
	WSADATA suWinsockData;

	// Start Winsock 2, if it fails, we do not need to call WSACleanup.
	wWinsockVersionRequested = MAKEWORD( 2, 2 );

	int nError = WSAStartup( wWinsockVersionRequested, &suWinsockData );

	if( nError )
	{
		// Failed to start winsock.
		MessageBox( HWND_DESKTOP,
    "Failed to start Winsock.",
    "Error",
    MB_ICONERROR );

		return false;
	}
	else
	{
		// Confirm that the Winsock 2 library supports the exact version we want if not, call WSACleanup.
		if( ( LOBYTE( suWinsockData.wVersion ) != 2 ) || ( HIBYTE( suWinsockData.wVersion ) != 2 ) )
		{
			WSACleanup();

			// Failed to get Winsock 2 library supports.
			DPRINT(0,"Failed to get Winsock 2 library supports.\n");

			return false;
		}
	}

	// Call WSAEnumProtocols to figure out how big of a buffer we need.
	DWORD dwBufferSize = 0;

	int nNumProtocols = WSAEnumProtocols( NULL, NULL, &dwBufferSize );

	if( ( nNumProtocols != SOCKET_ERROR ) && ( WSAGetLastError() != WSAENOBUFS ) )
	{
		WSACleanup();

		// Failed to get the number of protocols.
		DPRINT(0,"Failed to get the number of protocols.\n");

		return false;
	}

	// Allocate a buffer, call WSAEnumProtocols to get an array of WSAPROTOCOL_INFO structures.
	LPWSAPROTOCOL_INFO suSelectedProtocol = ( LPWSAPROTOCOL_INFO )malloc( dwBufferSize );

	if( suSelectedProtocol == NULL )
	{
		WSACleanup();

		// Failed to get the protocol informations.
		DPRINT(0,"Failed to get the protocol informations.\n");

		return false;
	}

	// Allocate memory for protocol list and define what protocols to look for.
	int* arnProtocols = ( int* )calloc( 2, sizeof( int ) );

	arnProtocols[ 0 ] = IPPROTO_TCP;
	arnProtocols[ 1 ] = IPPROTO_UDP;

	nNumProtocols = WSAEnumProtocols( arnProtocols, suSelectedProtocol, &dwBufferSize );

	free( arnProtocols );
	free( suSelectedProtocol );

	if( nNumProtocols == SOCKET_ERROR )
	{
    int myErr = WSAGetLastError();
		WSACleanup();

		// Failed to enumerate the protocols.
		DPRINT(0,"Failed to enumerate the protocols:%d.\n",myErr);

		return false;
	}
	DPRINT(0,"Inited wsocks\n");
  return true;

}
bool WinMessageManager::initListening(){
  int nError = 0;

	// Create the listening socket.
	mainSock = socket( AF_INET, SOCK_DGRAM, IPPROTO_UDP);

	if( mainSock == INVALID_SOCKET )
	{
		// Failed to create the socket.
		DPRINT(0,"Failed to create the socket.\n");

		return false;
	}

	// Allocate memory for the address structure and set it to zero.
	struct sockaddr*		psuSocketAddress;
	struct sockaddr_in*		psuSocketInternetAddress;

	psuSocketAddress	= ( struct sockaddr* )malloc( sizeof( sockaddr ) );

	memset( ( char* )psuSocketAddress, 0, sizeof( sockaddr ) );

	// Fill the address structure.
	psuSocketAddress->sa_family			= ( u_short )AF_INET;
	psuSocketInternetAddress			= ( struct sockaddr_in* )psuSocketAddress;
	psuSocketInternetAddress->sin_port	= htons( ( u_short )MAIN_PORT );

	// Bind the address information to the socket.
	nError = bind( mainSock, psuSocketAddress, sizeof( sockaddr ) );

	if( nError == SOCKET_ERROR )
	{
		free( psuSocketAddress );

		DPRINT(0,"Failed to bind the communication socket.\n");

		return false;
	}

	free( psuSocketAddress );
	//setsockopt
	// Everything's fine.
	DPRINT(0,"Inited binding\n");
	return true;

}

void WinMessageManager::basicSend(char *buf, int size, string addrIn, bool block ){
  struct sockaddr_in addr;
	getKeyed( addrIn, &addr );
	
  int nClientLength = sizeof(addr);
  if(block)
    outSockSem->enterCritical();
  if( sendto( mainSock, buf, size, 0, ( struct sockaddr* )&addr, nClientLength) == SOCKET_ERROR ){
    DPRINT(0,"failed to send invitation!!\n");
    if(block)
      outSockSem->exitCritical();
    return;
  }

  if(block)
    outSockSem->exitCritical();
	
	/*retrial rules*/
	/*
	reduntSem->enterCritical();
	
	reduntSem->exitCritical();
	*/
}
void WinMessageManager::sendLater(char *buf, int size, std::string addr, int deltaInMs){
	WinNodeInfo *wni = (WinNodeInfo *)myNI[addr];
	reduntSem->enterCritical();
	//TODO: retries should be default or node specific?
	wni->msgs2retry.push_back( new WinMessage(buf, size, deltaInMs,1) ); //FIXME: retries
	
	reduntSem->exitCritical();
}


int WinMessageManager::tryAndReceive(bool shortWait){
  /*
  */
	outSockSem->enterCritical();
	
  timeval timeout;
  if(shortWait){
    timeout.tv_sec = 0;
    timeout.tv_usec = SELECT_TO_US;
  } else {
    timeout.tv_sec = CONNECT_TO_S;
    timeout.tv_usec = 0;
  }

	

  fd_set rfds;
  FD_ZERO(&rfds);
  FD_SET(mainSock,&rfds);


  int result = select(0,&rfds,NULL, NULL, &timeout);
  if(result <= 0){
    //0 = timeout
    //-1 = error... but which?
    //cerr << "Timed out select:" << result << endl;
    outSockSem->exitCritical();
    //DPRINT(0,"\\\\Finished trying and receiving (TO)\n");
  
    return result;
  }

  // Initialize the internet socket address.
	struct sockaddr_in suInternetSocketAddress;

	int nClientLength = sizeof( suInternetSocketAddress );

	// Initialize the buffer.
	char szBuffer[ MAX_BUF ];


  // Read the incoming data from the communication socket.
  result = recvfrom( mainSock, szBuffer, MAX_BUF, 0, ( struct sockaddr* )&suInternetSocketAddress, &nClientLength );
  //TODO: run checksum
	//TODO: check for acks/retrying pkgs
  outSockSem->exitCritical();
	if( result != SOCKET_ERROR ){
		string mrReceived = getKey(suInternetSocketAddress);
		receivedPkg(szBuffer, mrReceived);
  
  }	else {
    //DPRINT(0,"Selected... but couldn't receive... something's not right\n");
    return -1;
  }

	
  //DPRINT(0,"\\\\Finished trying and receiving\n");
  return 0; //received right



}
string WinMessageManager::getKey(sockaddr_in sckin){
  char ipCStr[IP_CSTR_SIZE];
  strcpy(ipCStr,inet_ntoa(sckin.sin_addr));

  string key(ipCStr);
  
  return key;
}
void WinMessageManager::getKeyed(string in, sockaddr_in* out){
	out->sin_family = AF_INET;
	out->sin_port = htons( ( u_short )MAIN_PORT );
	out->sin_addr.s_addr = inet_addr(in.c_str());//returns const char*, no free needed
	
	if(out->sin_addr.s_addr == INADDR_NONE){
    DPRINT(0,"couldn't resolve \"%s\"\n", in.c_str());
  }
	
}


WinMessageManager::~WinMessageManager(){
	WSACleanup();
}
void WinMessageManager::store(string addrIn){
	DPRINT(0,"Storing %s", addrIn.c_str() );
	myNI[addrIn] = new WinNodeInfo();
	
	
}


void WinMessageManager::redundantLoop(){
	redundantLooping = true;
	while(redundantLooping){
		reduntSem->enterCritical();
		
		//int minimumTime = checkForQueued();
		checkForQueued();
		reduntSem->exitCritical();
		/*
		if(minimumTime < 0) //no new inputs
			reduntSem->waitForGreenLight(REDUNT_STEP);
		*/

		
		Sleep(REDUNDANT_STEP);

	}
}
void WinMessageManager::checkForQueued(){
	int msgsSent = 0;
	int cleared = 0;
	for(  map<string, P2pNodeInfo*>::iterator lIter = myNI.begin();
				lIter != myNI.end();
				lIter++){
		WinNodeInfo * currNode = (WinNodeInfo *)lIter->second;
		
		struct sockaddr_in addr;
		getKeyed( lIter->first, &addr );
		
		vector<WinMessage *>::iterator msgIter;
		for(	msgIter = currNode->msgs2retry.begin();
					msgIter != currNode->msgs2retry.end();
					msgIter++){
			WinMessage *currWMsg = *msgIter;
			if( !currWMsg->ready2send() )
				continue; //has something, but it's too soon to send
			
			int bufSize = currWMsg->getSize();
			//bufSize >= 0
			char *buf = new char[bufSize];
			if( bufSize == currWMsg->getContent( buf ) ){
				 
				outSockSem->enterCritical();
				sendto( mainSock, buf, bufSize, 0, ( struct sockaddr* )&addr, sizeof(addr));
				outSockSem->exitCritical();
				msgsSent++;
				
			}
			else{
				delete currWMsg;
				msgIter = currNode->msgs2retry.erase( msgIter );
				cleared++;
				
				
			}
			delete[] buf;
			//we could've deleted the last and can't operator++()
			if( msgIter == currNode->msgs2retry.end() )
					break;
			
		}
	}
	
	/*TODO: check if there's a missing confirming msg
	*/
	if(msgsSent > 0)
		DPRINT(0,"WMM:Resent %d msgs\n", msgsSent);
	if(cleared > 0)
		DPRINT(0,"WMM:Cleared %d msgs\n", cleared);
}

