#include "p2pSync.h"
#include "WinMessageManager.h"

#include <assert.h>

#ifdef __MINGW32__
  #define DPRINT if(FBA_DEBUG)bprintf
  extern int (__cdecl *bprintf) (int nStatus, TCHAR* szFormat, ...);
#else
  #define DPRINT if(FBA_DEBUG)bprintfPS
  void bprintfPS(int nStatus, TCHAR* szFormat, ...){ }
#endif

using namespace std;
 

#define INIT_TH 1 //coordinate after INIT_TH first sync
#define MIN_DELAY 3
#define MAX_STATES 3 //max states per player slot


int (WINAPI *gameCallback)(char *game, int player, int numplayers);

extern int clk2ms(clock_t in);


//is left equal or greater than right?

int msKey(int frameNo, int playerNo){
	return frameNo*100 + playerNo;
}

int msKeyedFrame(int key){
	return key/100;
	
}
int msKeyedPlayer(int key){
	return key%10;
}


p2pSync::p2pSync( struct p2pCallBacks* pCB){
  
	
	myGI = (struct gameInfo *)malloc( sizeof(struct gameInfo));
  
	p2pCB = (struct p2pCallBacks *)malloc( sizeof(struct p2pCallBacks) );
	p2pCB->stateSavePtr = pCB->stateSavePtr;
	p2pCB->stateLoadPtr = pCB->stateLoadPtr;
	p2pCB->pauseGamePtr = pCB->pauseGamePtr;
	p2pCB->unPauseGamePtr = pCB->unPauseGamePtr;
	if( 	!p2pCB->pauseGamePtr ||
				!p2pCB->unPauseGamePtr)
		DPRINT(0,"Wopa... esto no pasa!!!\n");
	if(   !p2pCB->stateSavePtr ||
				!p2pCB->stateLoadPtr){
		DPRINT(0,"Saving/loading disabled\n");
		p2pCB->stateSavePtr = NULL;
		p2pCB->stateLoadPtr = NULL;
	}
  bufSem = new P2pSem();

  //players = (P2PSplayer *)malloc(sizeof(P2PSplayer)*MAX_PLAYERS);
	for(int i=0;i<MAX_PLAYERS;i++)
    players.push_back( new P2pPlayer( i+1, myGI) );
    
	
  
	//lastFirstSync = (clock_t)-1;
	
}
void p2pSync::setData(list< string > lIpsIn,
                string gameNameIn, int playerNoIn,
                int totalPlayersIn, int gameIDIn){

  myGI->frameNo        = 0; //if you change this, you have to change it in P2pPlayer
	myGI->playerNo      = playerNoIn;
	myGI->gameID        = gameIDIn;

  myGI->firstSynced   = false;
	myGI->totalPlayers  = totalPlayersIn;
	myGI->inputSize     = -1;
	myGI->greenLighted = false;
	myGI->syncDelay  = 50;
	myGI->have2hold = 0;
	myGI->lastSync = clock();
	gameName = gameNameIn; 
	
	
  msgman = new WinMessageManager(lIpsIn,this);
	
   
}
int p2pSync::compareWithCurrent(int frameNo){
	return frameNo - myGI->frameNo;
}

int p2pSync::receivedData( unsigned char *inBuf, int inSize, struct gameInfo *inGameInfo){
	
	DPRINT(0,"p2p::rD, sync:%dp%d while @%u\n",
			inGameInfo->frameNo, inGameInfo->playerNo, myGI->frameNo);
	
	bufSem->enterCritical();
  int inFrameNo = inGameInfo->frameNo;
  int hisPlayerIndex=inGameInfo->playerNo-1;
	
	
	
	if( !(myGI->firstSynced) ){
		DPRINT(0,"...too bad we haven't started syncing\n");
		bufSem->exitCritical();
		return -1;
	}
	
	
	//if it's equal it's okay
	//myGI->frameNo points to the frame to be processed, not being proc'd

	assert( myGI->inputSize == inSize);
	//DPRINT(0,"myGI->inputSize:%d == inSize:%d\n",	myGI->inputSize, inSize);
	//assert( inFrameNo < (unsigned char)UCHAR_IDXS);

	int frameDiff = compareWithCurrent( inFrameNo );
	
  
	
  //Everything OK, copying to cache
  players[hisPlayerIndex]->setInput( (unsigned char *)inBuf, inSize, inFrameNo);
	
	if( frameDiff <= 0 ){
		
		int bounceFrameNo = myGI->frameNo+1;//process it on next "process", please
		
		//offer new bounce
		players[hisPlayerIndex]->bounceInput( inFrameNo, bounceFrameNo);
		
		//trick for stateSaving
		inFrameNo = bounceFrameNo;
    
		//return bounce frame
		inGameInfo->frameNo = bounceFrameNo;
	}
	

	queueStateSaving( inFrameNo, inGameInfo->playerNo);

  bufSem->exitCritical();
  
	return frameDiff;
}
void p2pSync::confirmBounce(int playerNo, int originalFrame, int bounceFrame){
	DPRINT(0,"ps: confBouncing p%d's %d to %d\n", playerNo, originalFrame, bounceFrame);
	//correct input if necessary
	bufSem->enterCritical();
	players[playerNo -1]->confirmBounce(originalFrame, bounceFrame);
	int key = msKey(originalFrame, playerNo);
	if( waitedInputs.size() > 0){//somebody was (probably) waiting for this?
		waitedInputs.erase( key );
		if( waitedInputs.size() == 0 )//we've just inputed the last waited
			bufSem->greenLight();
	}
	bufSem->exitCritical();
	
}
void p2pSync::receivedPlayerReady(struct gameInfo *inGameInfo){
  int hisPlayerI = inGameInfo->playerNo-1;
  
  players[hisPlayerI]->ready = true;
  DPRINT(0,"Player %d says he's ready @%u\n",
			inGameInfo->playerNo, clock() );
  
	
      
	if(myGI->playerNo != 1)
		DPRINT(0,"How come I recvd player ready and I'm not boss?\n");
	
	bool allReady=true;
	for(int i=0;i<myGI->totalPlayers;i++)
		if( !(players[i]->ready) ) //if one is not ready
			allReady=false; //none of us is
	if(allReady){
		DPRINT(0,"\nGO!(%u)\n", clock() );
		msgman->signalGo();
		bufSem->greenLight();
	}
}
void p2pSync::receivedGo(){
  
  myGI->greenLighted = true;
	bufSem->greenLight();
  
}


p2pSync::~p2pSync ()
{
	DPRINT(0,"deleting p2pSync Object\n");
  delete msgman;
  delete bufSem;
}

void p2pSync::greet(){
  msgman->greet();//send and receive greet
}

void p2pSync::chatSend(char *msg){

}
void p2pSync::sync(unsigned char *data, int size){
	//DPRINT(0,"sync:%d\n",myGI->frameNo+1);
	int pI = myGI->playerNo -1; //player index
  
  
  
  
  //this section is to give the emu some 'breathing' at drawing window
  

  if(!myGI->firstSynced){
		if(myGI->frameNo < INIT_TH){
			myGI->frameNo++;
			return;
		}
		
    
    firstSync(size);
  }
	
	if(myGI->have2hold){
		clock_t timeBefore = clock();
		int frames = myGI->have2hold/MS_PER_SYNC;
		if(myGI->have2hold%MS_PER_SYNC)
			frames++;
			
		//p2pCB->pauseGamePtr();
		Sleep(frames*MS_PER_SYNC);
		//p2pCB->unPauseGamePtr();
		DPRINT(0,"Holded for %ums\n",
			clk2ms(clock()-timeBefore));
		myGI->have2hold = 0;
	}
	
  //DPRINT(0,"%d=%d\n",myGI->inputSize,size);
  //assert(myGI->inputSize == size);
  
  //translate his input into our slot (paste it in the future)
  int futureFrameNo = myGI->frameNo +1 //frameNo == "last processed frame", s +1 = "frame in process"
																+ myGI->syncDelay;
  bool shouldBC = players[pI]->setInput( data, size, futureFrameNo);
	//bool shouldBC = players[pI]->compareWithOutCache( data, size );
 
  //assert(futureFrameNo < UCHAR_IDXS);
 
  
  //assert(size <= BUF_SIZE);
  
	if( shouldBC ){
		queueStateSaving( futureFrameNo, myGI->playerNo );
	}
	saveQueuedStates();
	loadQueuedStates();
	clearOldStates();
  //let's pack only our stuff for broadcast
  
   
  //broadcast this input
  if( shouldBC ){
	  clock_t bcastcp = clock();
    DPRINT(0,"Broadcasting new input from %d to %d\n", myGI->frameNo, futureFrameNo);
    msgman->broadcastData( data ,size, futureFrameNo);
		DPRINT(0,"Broadcasted took %dms\n",clk2ms(clock()-bcastcp) );
  }
  
	//check if all slots are ready
	bool allReady = false;
	int nextFrame = myGI->frameNo +1; //candidate for processing is the next one
	
	bufSem->enterCritical();

	while( !allReady ){
		int originalFrameNotReady = -1;
		allReady = true;
		for(int i=0; i< myGI->totalPlayers; i++){
			//bool printedOut = false;
			originalFrameNotReady = players[i]->inputReady(nextFrame);
			if( originalFrameNotReady >= 0){
				waitedInputs.insert( msKey(originalFrameNotReady, i+1) );
				DPRINT(0,"Input %up%d isn't (de)confirmed\n",
					nextFrame, i+1);
				allReady = false;
			}
		}
		if( !allReady ){
			DPRINT(0,"Waiting unconfirmed inputs on %d while @%d\n",
							nextFrame, myGI->frameNo);
			bufSem->exitCritical();
			bufSem->waitForGreenLight(16);
			bufSem->enterCritical();
			
			//originalFrameNotReady = players[i]->inputReady(myGI->frameNo);
		}
		
	}
	
  // all inputs ready... let's copy our slots to his
  int di = 0;
  unsigned char *input;
	input	= new unsigned char[size];
  for(int i=0;i < myGI->totalPlayers; i++){
    if( players[i]->getInput( (unsigned char *)input, nextFrame ) <= 0)
			DPRINT(0,"Read -1 from getInput at %u\n", nextFrame);
    for(int j=0;j<size;j++,di++)
      data[di] = input[j];
    
  }
	//break point
	myGI->frameNo++;//frame has been processed... move on
	
	delete[] input;

  //getReady for next loop
	
	
	//the following can be parallel, since we've used this frame
  bufSem->exitCritical();

  //this frame is clear
  for(int i=0;i<myGI->totalPlayers;i++){
    players[i]->clearInput( myGI->frameNo ); // "last processed frame is clear"
  }
  
	
}

void p2pSync::setGameCallback(int (WINAPI *gameCallbackIn)(char *game, int player, int numplayers)){

  gameCallback = gameCallbackIn;
  char gameNameC_STR[257];
  strcpy(gameNameC_STR,gameName.c_str());
  DPRINT(0,"p2p: Running \"%s\", I'm player %d/%d\n", gameNameC_STR,
      myGI->playerNo, myGI->totalPlayers);
  gameCallback(gameNameC_STR, myGI->playerNo, myGI->totalPlayers);
  DPRINT(0,"p2p: Callback ended\n");
}
void p2pSync::endGame(){
	
  DPRINT(0, "p2p: game should've ended\n");
}

void p2pSync::firstSync(int size){
  
  myGI->inputSize = size;
  int pI = myGI->playerNo -1;
  

	//msgman->getMaxDelay(); 


  
  players[pI]->ready = true; // I'm ready
  
	if( myGI->playerNo == 1){
		DPRINT(0,"Waiting for other players...");
		bufSem->waitForGreenLight(-1);
		msgman->signalGo();
  } else {
		//loop into signaling until we receive go
		while( bufSem->waitForGreenLight(2) && !myGI->greenLighted)// timeout and not green lighted in between
			msgman->signalReady(myGI);
		
	}
	
  myGI->firstSynced = true;
	DPRINT(0,"Exited first sync @%u (sync:%u)\n", clock(), myGI->frameNo );
  
  
  
}
void p2pSync::setNewDelay(int ms){
  
  if(ms < 0)
    return; //ignore
  
  bufSem->enterCritical();
	myGI->syncDelay  = ms/MS_PER_SYNC;
	
	if( ms%MS_PER_SYNC )
		myGI->syncDelay++;
	
	if(myGI->syncDelay < MIN_DELAY)
    myGI->syncDelay = MIN_DELAY; //testing found out that it was 2
  bufSem->exitCritical();
  DPRINT(0,"New sync delay:%d frames\n", myGI->syncDelay);


}
void p2pSync::deConfirmMsg(struct gameInfo *gi){
	DPRINT(0,"Deconfirming %dp%d\n", gi->frameNo, gi->playerNo);
	bufSem->enterCritical();
	
	players[ gi-> playerNo -1 ]->deConfirmMsg( gi->frameNo );
	deQueueStateSaving( gi->frameNo, gi->playerNo );
	
	if( waitedInputs.size() > 0){//somebody was waiting for this?
			waitedInputs.erase( msKey( gi->frameNo, gi->playerNo ) );
			if( waitedInputs.size() == 0 )//we've just inputed the last waited
				bufSem->greenLight();
	}
	bufSem->exitCritical();
	DPRINT(0,"Deconfirmed %up%d\n", gi->frameNo, gi->playerNo);
}

void p2pSync::confirmMsg( struct gameInfo *gi ){
	
	DPRINT(0,"p2p: confirmed %u while @%u\n", gi->frameNo, myGI->frameNo );
	int key = msKey( gi->frameNo, gi->playerNo );
	bufSem->enterCritical();
	if( gi->frameNo > myGI->frameNo
			|| waitedInputs.find( key ) != waitedInputs.end() ){
		players[ gi-> playerNo -1 ]->confirmMsg( gi->frameNo );
		
		if( waitedInputs.size() > 0){//somebody was waiting for this?
			waitedInputs.erase( key );
			if( waitedInputs.size() == 0 )//we've just inputed the last waited
				bufSem->greenLight();
		}
	}
	else {
		//broadcast state load
		unsigned char syncLoading = myGI->frameNo + myGI->syncDelay;
		DPRINT(0,"Queueing state loading from %u on %u\n", gi->frameNo, gi->frameNo + myGI->syncDelay );
		msgman->broadcastStateLoad( gi->frameNo, syncLoading, gi->playerNo);
		queueStateLoading( gi->frameNo, syncLoading, gi->playerNo);
		
		
	}
	bufSem->exitCritical();
	
}

void p2pSync::saveState(int futureFrameNo, int guiltyPlayerNo){
	if( !p2pCB->stateSavePtr ) //disabled
		return;
	int key = msKey( futureFrameNo, guiltyPlayerNo);
	
	mStates[key] = (struct machineState *)malloc(sizeof(machineState));
	mStates[key]->data = NULL;
	mStates[key]->size = 0;
	mStates[key]->frameNo = futureFrameNo;
	mStates[key]->guiltyPlayerNo = guiltyPlayerNo;
	mStates[key]->time2kill = clock()+ (clock_t)4*CLOCKS_PER_SEC;
	
	p2pCB->stateSavePtr( &(mStates[key]->data),&(mStates[key]->size) );
	DPRINT(0,"Saved state to %u\n", futureFrameNo);
	return;
}

void p2pSync::loadState(int frameNo, int playerNo){
	if( !p2pCB->stateLoadPtr )
		return;
	int key = msKey( frameNo,playerNo);
	if( mStates.find(key) == mStates.end() ){
		DPRINT(0,"Can't load state %d\n", key);
		return;
	}
	
	
	p2pCB->stateLoadPtr( mStates[key]->data, mStates[key]->size );
	/*
	free(mStates[key]->data);
	free(mStates[key]); //struct machineState
	mStates.erase(key);
	*/
}

void p2pSync::queueStateSaving(int frameNo, int playerNo){
	states2save.insert( msKey(frameNo, playerNo) );
}
void p2pSync::deQueueStateSaving(int frameNo, int playerNo){
	states2load.erase( msKey(frameNo, playerNo) );
}
void p2pSync::saveQueuedStates(){
	if( states2save.empty() )
		return;
	
	set<int>::const_iterator iter = states2save.begin();
	while(iter != states2save.end()){
		if( msKeyedFrame(*iter) == myGI->frameNo ){
			saveState( msKeyedFrame(*iter), msKeyedPlayer(*iter) );
			states2save.erase( iter++ );
		}	else
			++iter;
			
	}
}
void p2pSync::queueStateLoading(int frameSaved, int frameLoading, int playerNo){
	states2load[ frameLoading ] = msKey(frameSaved, playerNo);
}

void p2pSync::loadQueuedStates(){
	if( states2load.empty() )
		return;
	
	map<int, int>::iterator result =
		states2load.find(myGI->frameNo);
	if( result != states2load.end() ){
		int key = states2load[myGI->frameNo];
		loadState( msKeyedFrame(key), msKeyedPlayer(key) );
		states2load.erase(result);
	}
	
	
}



void p2pSync::clearOldStates(){ 
	if( states2load.empty() )
		return; //nothing to clear
	map<int, struct machineState *>::iterator iter = mStates.begin();
	while( iter != mStates.end() ){
		machineState *curr = iter->second;
		if( curr->time2kill >= clock() ){
			//destroy
			DPRINT(0,"Destroying mstate %d:%d\n", curr->frameNo, curr->guiltyPlayerNo);
			free( curr->data );
			free( curr ); //struct mState
			mStates.erase( iter++ );// first ++, then erase
		} else
			iter++;
		
	}
}

void p2pSync::hold(int ms){
	myGI->have2hold = ms;
}

struct gameInfo *p2pSync::getGI(){
	return myGI;
}

