#ifndef P2PSYNC
#define P2PSYNC
#include <time.h>
#define WIN32_LEAN_AND_MEAN
#ifndef _WIN32_WINNT
  #include <windows.h>
#endif

#include "CalleeBack.h"
#include "P2pSem.h"
#include "P2pPlayer.h"
#include "P2pMessage.h"

#include <string>
#include <list>
#include <map>
#include <vector>

#define MAX_PLAYERS 4
#define BUF_SIZE 96
#define SYNC_THRESHOLD 1
#define UCHAR_IDXS 256
#define CHAR_LOOP_TH (unsigned char)50
#define MS_PER_SYNC 16

struct gameInfo {
  int frameNo; //last frame processed, should be considered as past
  unsigned int playerNo; //1,2,3,4...
  unsigned int gameID;
	bool firstSynced;
	int totalPlayers;
	int inputSize;
	int syncDelay;
	unsigned int have2hold; //ms
	unsigned int frames2hold; //frames to ignore
	bool greenLighted;
	clock_t lastSync; //last clock() in which sync(...) was called
	
};

struct p2pCallBacks {
	void (__cdecl *stateSavePtr)(unsigned char**, unsigned int *);
	void (__cdecl *stateLoadPtr)(unsigned char *, unsigned int);
	void (__cdecl *pauseGamePtr)(void);
	void (__cdecl *unPauseGamePtr)(void);
		
};


struct machineState {
	unsigned char *data;
	unsigned int size;
	int frameNo;
	unsigned char guiltyPlayerNo;
	clock_t time2kill; //after this, this machineState is considered useless
};



class MessageManager;
class CalleeBack;
class p2pSync : public CalleeBack
{ 
  public:
		//std::list<std::string> lIPs, std::string gameName, int playerNo, int totalPlayers,int gameID
    
    ~p2pSync();
    int receivedData(unsigned char *, int, struct gameInfo *);
    void receivedPlayerReady(struct gameInfo *gameInfo);
    void receivedGo();
    void setNewDelay(int ms);
    void greet();//para testing
    void sync(unsigned char *buf,int size);
    void chatSend(char *msg);
    void endGame();
    void setGameCallback(int (WINAPI *gameCallbackIn)(char *game, int player, int numplayers));
    void confirmMsg( struct gameInfo *gi );
		void hold(int nFrames);
		struct gameInfo *getGI();
		void deConfirmMsg( struct gameInfo *gi );
		void confirmBounce(int playerNo, int originalFrame, int bounceFrame);
	public:
		void queueStateLoading( int syncSaved, int syncLoading, int playerNo);
		void queueStateSaving( int frameNo, int playerNo);
  protected:
		p2pSync( struct p2pCallBacks* pCB);
		int compareWithCurrent(int frameNo);
    void setData(std::list< std::string > lIps,
                std::string gameName, int playerNo,
                int totalPlayers, int gameID);
    void firstSync(int size);
		
    std::vector<P2pPlayer *> players;
    std::string gameName; //couldn't put it in gameInfo
    struct gameInfo *myGI;
    P2pSem *bufSem;
    MessageManager *msgman;
	protected:
		void saveQueuedStates(); //returns a reference to the state key
		void loadQueuedStates();
		void deQueueStateSaving( int frameNo, int playerNo);
		void saveState(int futureFrameNo, int playerNo);
		void loadState(int futureFrameNo, int playerNo);
		void clearOldStates();
		std::map<int, struct machineState *> mStates;
		std::set<int> states2save; //frameNo * 100 + playerNo ---> "reorders" itself
		std::set<int> waitedInputs;
		std::map<int, int> states2load; //frameNo * 100 + playerNo ---> "reorders" itself
		struct p2pCallBacks *p2pCB;
		
		//clock_t lastFirstSync;
};



#endif
